/**
 * Copyright (c) 2011, 2014, 2015, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_ARRAY_HPP_
#define INCLUDE_ARRAY_HPP_
#include <algorithm>
#include <cstddef>
#include <initializer_list>
#include <iterator>
#include <memory>
#include <stdexcept>
#include "overflow.hpp"

/* This class provides a simple fixed-size dynamic array; this array is not
 * intended to be a general-purpose replacement for std::vector; hence, you
 * should probably use std::vector. The only benefit of this class is that
 * arrays have a slightly smaller memory-footprint (which only matters if you
 * have many arrays). */
template<class Type>
class array final {
public:
    /* Container-style typedefs (interoperability with existing vector-code). */
    typedef Type value_type;
    typedef array<value_type> array_type;

	typedef typename std::size_t size_type;
	typedef typename std::ptrdiff_t difference_type;
	typedef value_type* pointer;
	typedef const value_type*  const_pointer;
	typedef value_type& reference;
	typedef const value_type&  const_reference;

    typedef pointer iterator;
    typedef const_pointer const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;


    /* Default-construct: empty array. */
    array() : data_pointer(nullptr), data_size(0) { }

    /* Construct array holding n copies of the specified value. */
    array(size_type n) : data_pointer(nullptr), data_size(0) {
        allocate_size(n);
        std::uninitialized_fill_n(data_pointer, n, value_type());
        data_size = n;
    }

    /* Construct array with the elements in the specified range. */
    template<class InputIterator>
    array(InputIterator first, InputIterator last) : data_pointer(nullptr), data_size(0) {
        size_type n = std::distance(first, last);
        allocate_size(n);
        std::uninitialized_copy(first, last, data_pointer);
        data_size = n;
    }

    /* initializer list constructor. */
    array(std::initializer_list<value_type> list) : array(list.begin(), list.end()) { }


    /* Copy-construct. */
    array(const array_type& other) : array(other.begin(), other.end()) { }

    /* Move-construct. */
    array(array_type&& other) : data_pointer(nullptr), data_size(0) {
        swap(other);
    }


    /* Destructor. */
    ~array() {
        if (std::is_trivially_destructible<value_type>::value) {
            auto it = data_pointer;
            auto end = data_pointer + data_size;
            for (; it != end; ++it) {
                it->~value_type();
            }
        }
        ::operator delete(data_pointer);
    }


    /* Copy-assignment: not provided. */
    array_type& operator=(const array_type& other) = delete;

    /* Move-assignment. */
    array_type& operator=(array_type&& other) {
        swap(other);
        return *this;
    }


    /* Return iterator to the begin of the array. */
    iterator begin() {
        return data_pointer;
    }
    const_iterator begin() const {
        return data_pointer;
    }
    const_iterator cbegin() const {
        return data_pointer;
    }

    /* Return iterator to the end of the array. */
    iterator end() {
        return data_pointer + data_size;
    }
    const_iterator end() const {
        return data_pointer + data_size;
    }
    const_iterator cend() const {
        return data_pointer + data_size;
    }


    /* Return reversed iterator to the end of the array. */
    reverse_iterator rbegin() {
        return reverse_iterator(end());
    }
    const_reverse_iterator rbegin() const {
        return const_reverse_iterator(end());
    }
    const_reverse_iterator crbegin() const {
        return const_reverse_iterator(end());
    }

    /* Return reversed iterator to the begin of the array. */
    reverse_iterator rend() {
        return reverse_iterator(begin());
    }
    const_reverse_iterator rend() const {
        return const_reverse_iterator(begin());
    }
    const_reverse_iterator crend() const {
        return const_reverse_iterator(begin());
    }


    /* Return the i-th element in the array. */
    reference operator[](size_type i) {
        return data_pointer[i];
    }
    const_reference operator[](size_type i) const {
        return data_pointer[i];
    }

    /* Return the i-th element in the array, will throw out_of_range when i not
     * in range. */
    reference at(size_type i) {
        if (!(0 <= i && i < data_size)) {
            throw std::out_of_range();
        }
        return data_pointer[i];
    }
    const_reference at(size_type i) const {
        if (!(0 <= i && i < data_size)) {
            throw std::out_of_range();
        }
        return data_pointer[i];
    }

    /* Return the size of the array. */
    size_type size() const {
        return data_size;
    }

    /* Swapping arrays. */
    void swap(array_type& other) {
        std::swap(other.data_pointer, data_pointer);
        std::swap(other.data_size, data_size);
    }

private:
    /* Allocate the raw storage for the array, throws bad_alloc when allocation
     * fails. */
    void allocate_size(const size_type n) {
        if (n > 0) {
            auto product = overflow_multiply<std::size_t>(n, sizeof(value_type));
            if (product.second) {
                throw std::bad_alloc();
            }
            data_pointer = (pointer) ::operator new(product.first);
        }
    }

    /* Pointer to dynamically allocated array and number of constructed elements
     * in the array. */
    pointer data_pointer;
    size_type data_size;
};

/* Return true if lhs < rhs, lexicographical ordering. */
template<class LhsType, class RhsType>
bool operator<(const array<LhsType>& lhs, const array<RhsType>& rhs) {
    return std::lexicographical_compare(lhs.begin(), lhs.end(), rhs.begin(), rhs.end());
}

/* Return true if lhs > rhs, lexicographical ordering. */
template<class LhsType, class RhsType>
bool operator>(const array<LhsType>& lhs, const array<RhsType>& rhs) {
    return rhs < lhs;
}

/* Return true if lhs <= rhs, lexicographical ordering. */
template<class LhsType, class RhsType>
bool operator<=(const array<LhsType>& lhs, const array<RhsType>& rhs) {
    return !(rhs < lhs);
}

/* Return true if lhs >= rhs, lexicographical ordering. */
template<class LhsType, class RhsType>
bool operator>=(const array<LhsType>& lhs, const array<RhsType>& rhs) {
    return !(lhs < rhs);
}

/* Return true if lhs == rhs, lexicographical ordering. */
template<class LhsType, class RhsType>
bool operator==(const array<LhsType>& lhs, const array<RhsType>& rhs) {
    return lhs.size() == rhs.size() && std::equal(lhs.begin(), lhs.end(), rhs.begin());
}

/* Return true if lhs != rhs, lexicographical ordering. */
template<class LhsType, class RhsType>
bool operator!=(const array<LhsType>& lhs, const array<RhsType>& rhs) {
    return !(lhs == rhs);
}

#endif