/**
 * Copyright (c) 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <fstream>
#include <iostream>
#include <string>
#include <stdexcept>
#include "array.hpp"
#include "bruteforce.hpp"

void help_message() {
    std::cout <<
        "bruteforce expressive power tool and benchmark\n"
        "    Usage: bfbenchclean scriptname\n"
        "           bfbenchclean scriptname --engine enginename\n"
        "           bfbenchclean scriptname --bench\n"
        "\n"
        "Run the bruteforce script specified by the scriptname. The script is either run\n"
        "with an automatically chosen engine, with the engine specified by enginename,\n"
        "or in benchmark mode (in which all suitable engines are used). Valid engines\n"
        "are the following: matrix8, matrix16, adjacency_bitset_list_32,\n"
        "adjacency_bitset_list_64,packed_edges_list_16_4, edge_list_8, edge_list_16,\n"
        "and edge_list_32.\n";
}

std::string read_file(const std::string& file_name) {
    std::ifstream source(file_name, std::ios::in | std::ios::binary);
    source.seekg(0, std::ios::end);
    if (!source.fail()) {
        auto pos = source.tellg();
        auto max = std::numeric_limits<std::string::size_type>::max();
        if (pos < 0) {
            throw std::runtime_error("could not read from input stream");
        }
        else if (static_cast<std::uintmax_t>(max) < static_cast<std::uintmax_t>(pos)) {
            throw std::runtime_error("input stream source to big");
        }
        else  {
            std::string contents;
            contents.resize(static_cast<std::string::size_type>(source.tellg()));
            if (contents.size() != 0) {
                source.seekg(0, std::ios::beg);
                source.read(&contents.front(), contents.size());
            }
            return contents;
        }
    }
    throw std::runtime_error("unable to read from input stream source");
}

int main(int argc, char* argv[]) {
    const std::string engine = "--engine";
    const std::string bench = "--bench";

    try {
        if (argc >= 2) {
            const std::string scriptname = argv[1];
            const std::string script = read_file(scriptname);
            auto scripts = read_scripts(script);

            if (argc == 2) {
                for (auto& script : scripts) {
                    auto engine = script->engine_auto();
                    script->execute(engine, std::cout, std::cerr);
                }
                return 0;
            }
            else if (argc == 3 && argv[2] == bench) {
                array<engine_type> engines = {
                    engine_type::matrix8,
                    engine_type::matrix16,
                    engine_type::adjacency_bitset_list_32,
                    engine_type::adjacency_bitset_list_64,
                    engine_type::packed_edges_list_16_4,
                    engine_type::edge_list_8,
                    engine_type::edge_list_16,
                    engine_type::edge_list_32
                };
                for (auto& script : scripts) {
                    for (auto& engine : engines) {
                        script->execute(engine, std::cout, std::cerr);
                    }
                }
                return 0;
            }
            else if (argc == 4 && argv[2] == engine) {
                std::string enginename = argv[3];
                auto engine = engine_string(enginename);
                for (auto& script : scripts) {
                    script->execute(engine, std::cout, std::cerr);
                }
                return 0;
            }
        }
        help_message();
    }
    catch (std::runtime_error exception) {
        std::cerr << "*** error during execution: " << exception.what() << ".\n";
    }
}