/**
 * Copyright (c) 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_BIT_ITERATOR_HPP_
#define INCLUDE_BIT_ITERATOR_HPP_
#include <type_traits>

/* The bitset implementations heavily relies on finding the finding the least
 * significant bit in an unsigned integer; for performance we use compiler
 * intrinsics to implement this. */
#include "intrin.h"

/* Iterator to iterate over the bits in an unsigned integer value. */
template<class UIntType>
class bit_iterator {
public:
    /* The unsigned integer type and the iterator type. */
    typedef UIntType uint_type;
    typedef bit_iterator<uint_type> iterator_type;
    static_assert(std::is_unsigned<uint_type>::value, "");

    /* Construct the bit iterator for the specified bitfield, offset each value
     * by the provided base value. */
    bit_iterator(const uint_type set, const uint_type base_value) :
                set(set), index(base_value) {
        if ((set & 1) == 0) {
            ++(*this);
        }
    }

    /* Construct the bit iterator for an empty value, valid end() iterator. */
    bit_iterator() : set(0), index(0) { }

    /* Return the index of the currently-set bit. */
    uint_type operator*() const {
        return index;
    }

    /* Move to the next-set bit. */
    void operator++() {
        set >>= 1;
        auto shift_index = bit_scan_forward(set);
        set >>= shift_index;
        index += shift_index + 1;
    }

    /* Non-equality. Will evaluate to true when not-both sets are empty (end())
     * or when the two iterators iterate over the same set but are at different
     * indices (in other cases; when iterators are over different sets; the
     * behavior is undefined). */
    bool operator!=(iterator_type& other) {
        return other.set != set;
    }

private:
    /* Return the first set bit (lkeast significant bit). */
    static uint_type bit_scan_forward(const uint_type mask) {
        unsigned long index;
        _BitScanForward64(&index, mask);
        return static_cast<uint_type>(index);
    }

    /* The set and the current index in the set. */
    uint_type set;
    uint_type index;
};

/* Bit iterator `container'. */
template<class UIntType>
class bit_iterator_container {
public:
    typedef UIntType uint_type;

    bit_iterator_container(const uint_type set, const uint_type base_value) :
                set(set), index(base_value) { }

    bit_iterator<uint_type> begin() const {
        return bit_iterator<uint_type>(set, index);
    }

    bit_iterator<uint_type> end() const {
        return bit_iterator<uint_type>();
    }

private:
    /* The set and the current index in the set. */
    uint_type set;
    uint_type index;
};

/* Construct and return the bit iterator container for the specified value. */
template<class UIntType>
bit_iterator_container<UIntType> make_bit_iterator(const UIntType value) {
    return bit_iterator_container<UIntType>(value, 0);
}

template<class UIntType>
bit_iterator_container<UIntType> make_bit_iterator(const UIntType value, const UIntType base_value) {
    return bit_iterator_container<UIntType>(value, base_value);
}

#endif