/**
 * Copyright (c) 2014, 2015, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <algorithm>
#include <chrono>
#include <iostream>
#include <map>
#include "array.hpp"
#include "bruteforce.hpp"

#define INCLUDE_BRUTEFORCE_CPP_DEPENDENCY
    #include "bruteforce_relational.hpp"
    #include "bruteforce_compute.hpp"
#undef INCLUDE_BRUTEFORCE_CPP_DEPENDENCY

constexpr char prepend_no_prev[] = "";
constexpr char prepend_prev[] = ", ";

/**
 * Write a relation to standard output, using the specified node names as the
 * name of the node identifiers used in the relation.
 */
template<class RelationType>
void relation_output(const RelationType& relation,
                     const std::vector<std::string>& node_names,
                     std::ostream& output) {
    raw_relation raw_relation = relation.to_relation();
    const char* prepend = prepend_no_prev;
    for (auto& edge : raw_relation) {
        output << prepend << '(' << node_names[edge.from] << ','
                                 << node_names[edge.to] << ')';
        prepend = prepend_prev;
    }
}

/**
 * Print the results for bruteforcing separability for path queries given a pair
 * of languages (indicesd in the script configurations and the result sets).
 */
template<class ResultsetType>
bool print_case(std::size_t i, std::size_t j, const path_script& script,
                const ResultsetType& result_sets,
                std::ostream& output) {
    output << "    language " << script.configurations[i] << " and "
                              << script.configurations[j] << ":\n";
    auto it_i = result_sets[i].entry_set.begin();
    auto end_i = result_sets[i].entry_set.end();
    auto it_j = result_sets[j].entry_set.begin();
    auto end_j = result_sets[j].entry_set.end();
    while (it_i != end_i) {
        while (it_j != end_j && it_j->first < it_i->first) {
            ++it_j;
        }
        if (it_j == end_j || it_i->first < it_j->first) {
            output << "        separating query Q: ";
            it_i->second.to_output(result_sets[i].entry_list, script.graph.relation_names, output);
            output << "\n        Q(" << script.graph.name << ") = {";
            relation_output(it_i->first, script.graph.node_names, output);
            output << "}\n\n";

            return true;
        }
        else {
            ++it_i;
        }
    }
    return false;
}

std::ostream& operator<<(std::ostream& output, const engine_type& engine) {
    switch (engine) {
        #define WRITE_CASE(NAME)   \
            case engine_type::NAME:    \
                return output << #NAME;

            WRITE_CASE(matrix8)
            WRITE_CASE(matrix16)
            WRITE_CASE(adjacency_bitset_list_32)
            WRITE_CASE(adjacency_bitset_list_64)
            WRITE_CASE(packed_edges_list_16_4)
            WRITE_CASE(edge_list_8)
            WRITE_CASE(edge_list_16)
            WRITE_CASE(edge_list_32)
            WRITE_CASE(invalid_engine)

        #undef WRITE_CASE


        default:
            return output << "invalid engine: default";
    }
}

engine_type engine_string(const std::string& engine_name) {
    static const std::map<std::string, engine_type> values = {
        #define WRITE_PAIR(NAME) \
            { #NAME, engine_type::NAME }

            WRITE_PAIR(matrix8),
            WRITE_PAIR(matrix16),
            WRITE_PAIR(adjacency_bitset_list_32),
            WRITE_PAIR(adjacency_bitset_list_64),
            WRITE_PAIR(packed_edges_list_16_4),
            WRITE_PAIR(edge_list_8),
            WRITE_PAIR(edge_list_16),
            WRITE_PAIR(edge_list_32)

        #undef WRITE_PAIR
    };

    auto it = values.find(engine_name);
    return (it != values.end()) ? it->second : engine_type::invalid_engine;
}

engine_type engine_fit_case(const std::size_t max_num_nodes) {
    static const array<std::pair<std::size_t, engine_type>> values = {
        #define WRITE_PAIR(NAME) \
            { NAME::max_num_nodes, engine_type::NAME }

            WRITE_PAIR(matrix8),
            WRITE_PAIR(matrix16),
            WRITE_PAIR(adjacency_bitset_list_32),
            WRITE_PAIR(adjacency_bitset_list_64),
            WRITE_PAIR(packed_edges_list_16_4),
            WRITE_PAIR(edge_list_8),
            WRITE_PAIR(edge_list_16),
            WRITE_PAIR(edge_list_32)

        #undef WRITE_PAIR
    };

    for (auto& pair : values) {
        if (max_num_nodes <= pair.first) {
            return pair.second;
        }
    }
    return engine_type::invalid_engine;
}

std::ostream& operator<<(std::ostream& output, const configuration& configuration) {
    const array<std::pair<bool, std::string>> values = {
            { configuration.diversity,  "di" },
            { configuration.converse, "~"  },
            { configuration.project,  "project" },
            { configuration.coproject, "coproject" },
            { configuration.intersect, "intersect" },
            { configuration.difference, "difference" }
    };

    output << "N(";
    const char* prepend = prepend_no_prev;
    for (auto& pair : values) {
        if (pair.first) {
            output << prepend << pair.second;
            prepend = prepend_prev;
        }
    }
    output << ")";

    return output;
}


boolean_script::relation_id_pairs_list
boolean_script::create_relation_id_pairs(const graph& first_graph, const graph& second_graph) {
    std::map<std::string, std::pair<relation_id_type, relation_id_type>> mapping;
    for (relation_id_type i = 0; i < first_graph.num_relations(); ++i) {
        auto& name = first_graph.relation_names[i];
        mapping.emplace(name, std::make_pair(i, second_graph.num_relations()));
    }
    for (relation_id_type i = 0; i < second_graph.num_relations(); ++i) {
        auto& name = second_graph.relation_names[i];
        auto it = mapping.emplace(name, std::make_pair(first_graph.num_relations(), i));
        if (!it.second) {
            it.first->second.second = i;
        }
    }

    relation_id_pairs_list result;
    for (const auto& value : mapping) {
        result.emplace_back(value.second);
    }
    return result;
}

std::vector<std::string>
boolean_script::create_relation_names(const graph& first_graph, const graph& second_graph,
                                      const relation_id_pairs_list& relation_id_pairs) {
    std::vector<std::string> result;
    for (auto& pair : relation_id_pairs) {
        if (pair.first != first_graph.num_relations()) {
            result.emplace_back(first_graph.relation_names[pair.first]);
        }
        else {
            result.emplace_back(second_graph.relation_names[pair.second]);
        }
    }
    return result;
}

template<class RelationType>
void execute_path(const path_script& script, std::ostream& output) {
    output << "path separability for `" << script.name << "'\n";
    std::vector<construct_result_path<RelationType>> result_sets;
    for (auto& configuration : script.configurations) {
        construct_result_path<RelationType> result(script, configuration);
        bruteforce_compute(result, configuration);
        result_sets.emplace_back(std::move(result));
    }

    for (std::size_t i = 0; i < script.configurations.size(); ++i) {
        for (std::size_t j = 0; j < script.configurations.size(); ++j) {
            if (i != j) {
                if (!print_case(i, j, script, result_sets, output)) {
                    output << "        no separating query found\n\n";
                }
            }
        }
    }
    output << '\n';
}

void path_script::execute(const engine_type engine, std::ostream& output, std::ostream& benchlog) const {
    using namespace std::chrono;
    auto before = high_resolution_clock::now();
    switch (engine) {
        #define WRITE_CASE(NAME)                   \
            case engine_type::NAME:                \
                execute_path<NAME>(*this, output); \
                break;

            WRITE_CASE(matrix8)
            WRITE_CASE(matrix16)
            WRITE_CASE(adjacency_bitset_list_32)
            WRITE_CASE(adjacency_bitset_list_64)
            WRITE_CASE(packed_edges_list_16_4)
            WRITE_CASE(edge_list_8)
            WRITE_CASE(edge_list_16)
            WRITE_CASE(edge_list_32)

        #undef WRITE_CASE

        case engine_type::invalid_engine:
        default:
            throw std::invalid_argument("engine not provided or invalid engine");
            break;
    }
    auto after = high_resolution_clock::now();
    benchlog << name << " " << engine << " " << duration_cast<milliseconds>(after - before).count() << "ms\n";
}

engine_type path_script::engine_auto() const {
    return engine_fit_case(graph.num_nodes());
}

template<class RelationType>
void execute_boolean(const boolean_script& script, std::ostream& output) {
    output << "boolean separability for `" << script.name << "'\n";
    for (auto& configuration : script.configurations) {
        output << "    language " << configuration << ":\n";
        construct_result_boolean<RelationType> result(script, configuration);
        try {
            bruteforce_compute(result, configuration);
            std::cout << "        no separating query found\n\n";
        }
        catch (const boolean_separation_exception separation_result) {
            auto& entry = *result.entry_list[separation_result.entry_id];

            output << "        separating query Q: ";
            entry.second.to_output(result.entry_list, script.relation_names, output);
            output << "\n        Q(" << script.first_graph.name << ") = {";
            relation_output(entry.first.first, script.first_graph.node_names, output);
            output << "}\n        Q(" << script.second_graph.name << ") = {";
            relation_output(entry.first.second, script.second_graph.node_names, output);
            output << "}\n\n";
        }
    }
}

void boolean_script::execute(const engine_type engine, std::ostream& output, std::ostream& benchlog) const {
    using namespace std::chrono;
    auto before = high_resolution_clock::now();
    switch (engine) {
        #define WRITE_CASE(NAME)                      \
            case engine_type::NAME:                   \
                execute_boolean<NAME>(*this, output); \
                break;

            WRITE_CASE(matrix8)
            WRITE_CASE(matrix16)
            WRITE_CASE(adjacency_bitset_list_32)
            WRITE_CASE(adjacency_bitset_list_64)
            WRITE_CASE(packed_edges_list_16_4)
            WRITE_CASE(edge_list_8)
            WRITE_CASE(edge_list_16)
            WRITE_CASE(edge_list_32)

        #undef WRITE_CASE

        case engine_type::invalid_engine:
        default:
            throw std::invalid_argument("engine not provided or invalid engine");
            break;
    }
    auto after = high_resolution_clock::now();
    benchlog << name << " " << engine << " " << duration_cast<milliseconds>(after - before).count() << "ms\n";
}

engine_type boolean_script::engine_auto() const {
    return engine_fit_case(std::max(first_graph.num_nodes(), second_graph.num_nodes()));
}