/**
 * Copyright (c) 2013, 2014, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_BRUTEFORCE_HPP_
#define INCLUDE_BRUTEFORCE_HPP_
#include <cstdint>
#include <iosfwd>
#include <memory>
#include <string>
#include <utility>
#include <vector>
#include "graph.hpp"

/* Possible relation-engine types. */
enum class engine_type {
    matrix8,
    matrix16,
    adjacency_bitset_list_32,
    adjacency_bitset_list_64,
    packed_edges_list_16_4,
    edge_list_8,
    edge_list_16,
    edge_list_32,
    invalid_engine
};

std::ostream& operator<<(std::ostream& output, const engine_type& engine);

/* Return the best-fit relation-engine for the specified use case. */
engine_type engine_fit_case(const std::size_t max_num_nodes);

/* Return the relation-engine with the specified name. */
engine_type engine_string(const std::string& engine_name);

/* Language configuration; by default all extra features are disabled. */
struct configuration {
    bool diversity = false;
    bool converse = false;
    bool project = false;
    bool coproject = false;
    bool intersect = false;
    bool difference = false;
};

std::ostream& operator<<(std::ostream& output, const configuration& configuration);

/* Bruteforce script. */
class script {
public:
    script(const std::string& name, const std::vector<configuration>& configurations) : 
            name(name), configurations(configurations) { }

    virtual ~script() { }

    /* Execute the script. */
    virtual void execute(const engine_type engine, std::ostream& output, std::ostream& benchlog) const = 0;

    /* Select a fitting engine for this script. */
    virtual engine_type engine_auto() const = 0;

    /* Name of the script. */
    const std::string name;

    /* Language configurations for which the script should be executed. */
    const std::vector<configuration> configurations;

};

/* A bruteforce script for determining path distinguishability. */
class path_script : public script {
public:
    path_script(const std::string& name, const std::vector<configuration>& configurations, const graph& graph) :
            script(name, configurations), graph(graph) { }

    void execute(const engine_type engine, std::ostream& output, std::ostream& benchlog) const;
    engine_type engine_auto() const;

    /* The graph to compute all possible query outputs for. */
    const graph graph;
};

/* A bruteforce script for determining boolean distinguishability. */
class boolean_script : public script {
public:
    boolean_script(const std::string& name, const std::vector<configuration>& configurations,
                    const graph& first_graph, const graph& second_graph) :
            script(name, configurations),
            first_graph(first_graph), second_graph(second_graph),
            relation_id_pairs(create_relation_id_pairs(first_graph, second_graph)),
            relation_names(create_relation_names(first_graph, second_graph, relation_id_pairs)) { }


    void execute(const engine_type engine, std::ostream& output, std::ostream& benchlog) const;
    engine_type engine_auto() const;

    /* The graphs to try to distinguish. */
    const graph first_graph;
    const graph second_graph;

    /* The graphs might have different mappings between relation names and
     * relation identifiers. Hence, we maintain a list of pairs of relation
     * identifiers used in the first and the second graph in such a way that the
     * identifiers refer to the same relation names. For those relations that
     * only exist in one of the graphs, it will use a value bigger than any of
     * the identifiers in use (for the respective graph). */
    typedef std::vector<std::pair<relation_id_type, relation_id_type>> relation_id_pairs_list;
    const relation_id_pairs_list relation_id_pairs;
    const std::vector<std::string> relation_names;

private:
    /* Construct the relation id pair structure. */
    static relation_id_pairs_list create_relation_id_pairs(const graph& first_graph,
                                                           const graph& second_graph);

    /* Construct a relation name list for the indices in the relation id pairs list. */
    static std::vector<std::string> create_relation_names(const graph& first_graph,
                                                          const graph& second_graph,
                                                          const relation_id_pairs_list& relation_id_pairs);
};

/* Read a list of scripts from the specified input. */
std::vector<std::unique_ptr<script>> read_scripts(const std::string& source);

#endif