/**
 * Copyright (c) 2014, 2015, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_BRUTEFORCE_COMPUTE_HPP_
#define INCLUDE_BRUTEFORCE_COMPUTE_HPP_
#include <algorithm>
#include "bruteforce_tree.hpp"

#ifndef INCLUDE_BRUTEFORCE_CPP_DEPENDENCY
    #error "not intended for direct inclusion, include bruteforce.hpp instead"
#endif


template<class RelationType>
void debug_output(const RelationType& relation) {
    constexpr char prepend_no_prev[] = "";
    constexpr char prepend_prev[] = ", ";
    raw_relation raw_relation = relation.to_relation();
    const char* prepend = prepend_no_prev;
    for (auto& edge : raw_relation) {
        std::cout << prepend << '(' << edge.from << ','
                                    << edge.to << ')';
        prepend = prepend_prev;
    }
}

/* Performing relational operators, using a buffer. */
template<class RelationType, class BufferType>
class perform_operation {
public:
    typedef RelationType relation_type;
    typedef BufferType buffer_type;
 
    /* Construct this operation performer by initializing the buffer. */
    perform_operation(const std::size_t max_num_nodes) :
            buffer(relation_type::initial_buffer(max_num_nodes)) { }

    /* Move-construct this operation performer. */
    perform_operation(perform_operation<relation_type, buffer_type>&& other) :
            buffer(std::move(other.buffer)) { }

    /* Perform the specified operation. */
    template<class Operation, class... Args>
    relation_type perform(const Args&... relations) {
        return Operation::perform(relations..., buffer);
    }

    /* Perform the specified construction operation (identity, diversity, and
     * constructing relations). */
    template<class Operation, class Arg>
    relation_type construct(const Arg& argument) {
        return Operation::construct<RelationType>(argument, buffer);
    }

private:
    /* The buffer to use. */
    buffer_type buffer;
};

/* Performing relational operators, not using a buffer. */
template<class RelationType>
class perform_operation<RelationType, relation::relation_inplace_tag> {
public:
    typedef RelationType relation_type;

    /* Construct this operation performer. */
    perform_operation(const std::size_t num_nodes) { }

    /* Move-construct this operation performer. */
    perform_operation(perform_operation<relation_type, std::false_type>&& other) { }

    /* Perform the specified operation. */
    template<class Operation, class... Args>
    relation_type perform(const Args&... relations) {
        return Operation::perform(relations...);
    }

    /* Perform the specified construction operation (identity, diversity, and
     * constructing relations). */
    template<class Operation, class Arg>
    relation_type construct(const Arg& argument) {
        return Operation::construct<RelationType>(argument);
    }
};

/* Thrown when a boolean separation result has been found.  The entry id
 * represents the index of the entry in the entry list in the constructed result
 * set that proves the separation. */
struct boolean_separation_exception {
public:
    const std::size_t entry_id;
};

/* The construct result set implements how an operation is executed with respect
 * to the script type. To determine boolean separation (via boolean scripts)
 * query operations need to be applied on pairs of relations, whereas to
 * determine  path separation (via path scripts) query operations need to be
 * applied on a single relation. Overloading core functions and the constructors
 * with respect to the differences in boolean and path scripts implements these
 * differences in bruteforce details. */
template<class RelationType, bool boolean>
class construct_result_set {
public:
    typedef RelationType relation_type;
    typedef std::pair<relation_type, relation_type> relation_pair_type;
    typedef typename std::conditional<boolean, relation_pair_type, relation_type>::type key_type;
    typedef typename std::conditional<boolean, boolean_script, path_script>::type script_type;
    typedef construct_result_set<relation_type, boolean> result_type;

    typedef bruteforce_tree tree_node_type;
    typedef std::map<key_type, tree_node_type> entry_set_type;
    typedef typename entry_set_type::value_type entry_type;
    typedef typename entry_set_type::const_iterator entry_iterator;
    typedef std::vector<entry_iterator> entry_list_type;


    /* We provide no copy or assignment operators. */
    construct_result_set(const result_type& other) = delete;
    result_type& operator=(const result_type& other) = delete;
    result_type& operator=(result_type&& other) = delete;

    /* Construct the result set for a specified graph and configuration,
     * this constructor is called for boolean scripts. */
    construct_result_set(const boolean_script& script, const configuration configuration) :
            script(script), entry_set(), entry_list(),
            operation_performer(std::max(script.first_graph.num_nodes(),
                                         script.second_graph.num_nodes())),
            identity(operation_performer.construct<op_identity>(script.first_graph.num_nodes()),
                     operation_performer.construct<op_identity>(script.second_graph.num_nodes())) { }

    /* Construct the result set for a specified graph and configuration,
     * this constructor is called for path scripts. */
    construct_result_set(const path_script& script, const configuration configuration) :
            script(script), entry_set(), entry_list(),
            operation_performer(script.graph.num_nodes()),
            identity(operation_performer.construct<op_identity>(script.graph.num_nodes())) { }

    /* Move-constructor. */
    construct_result_set(result_type&& other) :
            script(other.script), 
            entry_set(std::move(other.entry_set)),
            entry_list(std::move(other.entry_list)),
            identity(std::move(other.identity)),
            operation_performer(std::move(other.operation_performer)) { }


    /* The script used for this task. */
    const script_type& script;

    /* The entry set is a set of (relation, expression)-pairs. The relation are
     * query results together with the expression (as an expression tree) that
     * resulted in this query result. The entry_list provides iterators to these
     * (relation, expression)-pairs and, hence, provides a mapping from
     * (relation, expression)-pair-identifiers to the entry in the entry set.
     * These identifiers are used in the expression tree (as pointers). */
    entry_set_type entry_set;
    entry_list_type entry_list;

    /* Used to execute relational operators while taking into account any
     * specific details of the relation implementation (possible providing
     * buffers to use by the relational operators). */
    perform_operation<relation_type, typename relation_type::buffer_type> operation_performer;

    /* The identity relation. */
    const key_type identity;

    /* Add the constants and the relations to the entry set, return indices to
     * the first and one-past-last elements in the entry list that represent
     * relations. */
    std::pair<std::size_t, std::size_t> add_initials(const configuration configuration) {
        return add_initials(script, configuration);
    }

    /* The generate functions generate a relation by performing the operation
     * Operation (unary of binary) using the specified arguments (either
     * relations, for path scripts, or relation pairs, for boolean scripts).
     * These method takes care of buffers whenever buffers are necessary. */
     
    /* Path scripts: unary operations on relations. */
    template<class Operation>
    relation_type generate(const relation_type& relation) {
        return operation_performer.perform<Operation>(relation);
    }

    /* Path scripts: binary operations on relations. */
    template<class Operation>
    relation_type generate(const relation_type& left, const relation_type& right) {
        return operation_performer.perform<Operation>(left, right);
    }

    /* Boolean scripts: unary operations on relations. */
    template<class Operation>
    relation_pair_type generate(const relation_pair_type& pair) {
        return std::make_pair(operation_performer.perform<Operation>(pair.first),
                              operation_performer.perform<Operation>(pair.second));
    }

    /* Boolean scripts: binary operations on relations. */
    template<class Operation>
    relation_pair_type generate(const relation_pair_type& left, const relation_pair_type& right) {
        return std::make_pair(operation_performer.perform<Operation>(left.first, right.first),
                              operation_performer.perform<Operation>(left.second, right.second));
    }
    

    /* Check the generated query results: for determining boolean separation, we
     * can stop whenever the relation pair consists of an empty and a non-empty
     * relation. For determining path separation, we need to compute all
     * relations (no checks needed). */
    void add_check(const relation_type& relation, const std::size_t entry_id) {
    }

    void add_check(const relation_pair_type& pair, const std::size_t entry_id) {
        if (pair.first.is_empty() != pair.second.is_empty()) {
            throw boolean_separation_exception{entry_id};
        }
    }
    
    /* Add a generated key with the specified details (for the expression tree
     * node) to the entry set. */
    template<class... Args>
    void add(key_type&& key, const Args&... args) {
        auto r = entry_set.emplace(std::move(key), tree_node_type(args...));
        if (r.second) {
            entry_list.emplace_back(r.first);
            add_check(r.first->first, entry_list.size() - 1);
        }
    }

    /* Generate the relation by executing the specified unary operation and then
     * add the relation to the entry set. */
    template<class Operation>
    void add_generate(const operation_type type,
                      const entry_type& single, const std::size_t single_id) {
        add(generate<Operation>(single.first), type, single_id);
    }

    /* Generate the relation by executing the specified binary operation and
     * then add the relation to the entry set. */
    template<class Operation, class... Args>
    void add_generate(const operation_type type,
                      const entry_type& lhs, const entry_type& rhs,
                      const std::size_t lhs_id, const std::size_t rhs_id) {
        add(generate<Operation>(lhs.first, rhs.first), type, lhs_id, rhs_id);
    }


private:
    /* Add the initial constants and relations (path scripts). */
    std::pair<std::size_t, std::size_t> add_initials(const path_script& script, const configuration configuration) {
        /* Add constants. */
        add(relation_type(identity), operation_type::identity);
        if (configuration.diversity) {
            add(operation_performer.construct<op_diversity>(script.graph.num_nodes()), operation_type::diversity);
        }

        /* Add the relations and set the relation range. */
        std::size_t first = entry_list.size();
        for (std::size_t id = 0; id < script.graph.num_relations(); ++id) {
            add(operation_performer.construct<op_relation>(script.graph.relations[id]), id, false);
        }
        return std::make_pair(first, entry_list.size());
    }

    /* Add the initial constants and relations (boolean scripts). */
    std::pair<std::size_t, std::size_t> add_initials(const boolean_script& script, const configuration configuration) {
        /* Add constants. */
        add(key_type(identity), operation_type::identity);
        if (configuration.diversity) {
            add(std::make_pair(operation_performer.construct<op_diversity>(script.first_graph.num_nodes()),
                               operation_performer.construct<op_diversity>(script.second_graph.num_nodes())),
                operation_type::diversity);
        }

        /* Add the relations and set the relation range. */
        std::size_t first = entry_list.size();
        auto& id_pairs = script.relation_id_pairs;
        auto& first_graph = script.first_graph;
        auto& second_graph = script.second_graph;
        for (std::size_t id = 0; id < id_pairs.size(); ++id) {
            auto first_relation = (id_pairs[id].first != first_graph.num_relations()) ?
                                  operation_performer.construct<op_relation>(first_graph.relations[id_pairs[id].first]) : relation_type();
            auto second_relation = (id_pairs[id].second != second_graph.num_relations()) ?
                                  operation_performer.construct<op_relation>(second_graph.relations[id_pairs[id].second]) : relation_type();
            add(std::make_pair(std::move(first_relation), std::move(second_relation)), id, false);
        }
        return std::make_pair(first, entry_list.size());
    }
};

/* The result set constructor for path scripts. */
template<class RelationType>
using construct_result_path = construct_result_set<RelationType, false>;


/* The result set constructor for booolean scripts. */
template<class RelationType>
using construct_result_boolean = construct_result_set<RelationType, true>;


/* The main bruteforce function: provided a result set constructor that already
 * contains all constants and relations; compute all* possible query results
 * obtainable by queries written in the provided configuration. The precise
 * definition of a ``query result'' depends on the query strategies provided by
 * the result set constructor. */
template<class RelationType, bool boolean>
void bruteforce_compute(construct_result_set<RelationType, boolean>& result,
                        const configuration configuration) {
    auto pair = result.add_initials(configuration);
    if (configuration.converse) {
        for (std::size_t entry_id = pair.first; entry_id < pair.second; ++entry_id) {
            auto& entry = *result.entry_list[entry_id];
            result.add_generate<op_converse>(operation_type::converse_relation,
                                             entry, entry.second.relation.relation_id);
        }
    }

    for (std::size_t entry_id = 0; entry_id < result.entry_list.size(); ++entry_id) {
        auto& entry = *result.entry_list[entry_id];

        result.add_generate<op_compose>(operation_type::relation_compose,
                                        entry, entry, entry_id, entry_id);

        if (configuration.coproject) {
            auto project_1 = result.generate<op_project_1>(entry.first);
            auto project_2 = result.generate<op_project_2>(entry.first);
            auto coproject_1 = result.generate<op_difference>(result.identity, project_1);
            auto coproject_2 = result.generate<op_difference>(result.identity, project_2);
            if (configuration.project) {
                result.add(std::move(project_1), operation_type::project_1, entry_id);
                result.add(std::move(project_2), operation_type::project_2, entry_id);
            }
            result.add(std::move(coproject_1), operation_type::coproject_1, entry_id);
            result.add(std::move(coproject_1), operation_type::coproject_2, entry_id);
        }
        else if (configuration.project) {
            result.add_generate<op_project_1>(operation_type::project_1, entry, entry_id);
            result.add_generate<op_project_2>(operation_type::project_2, entry, entry_id);
        }

        for (std::size_t other_id = 0; other_id < entry_id; ++other_id) {
            auto& other = *result.entry_list[other_id];
            result.add_generate<op_compose>(operation_type::relation_compose,
                                            entry, other, entry_id, other_id);
            result.add_generate<op_compose>(operation_type::relation_compose,
                                            other, entry, other_id, entry_id);
            result.add_generate<op_union>(operation_type::relation_union,
                                          other, entry, entry_id, other_id);

            if (configuration.intersect) {
                result.add_generate<op_intersect>(operation_type::relation_intersect,
                                                  entry, other, entry_id, other_id);
            }
            if (configuration.difference) {
                result.add_generate<op_difference>(operation_type::relation_difference,
                                                   entry, other, entry_id, other_id);
                result.add_generate<op_difference>(operation_type::relation_difference,
                                                   other, entry, other_id, entry_id);
            }
        }
    }
}

#endif