/**
 * Copyright (c) 2014, 2015, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_BRUTEFORCE_RELATIONAL_HPP_
#define INCLUDE_BRUTEFORCE_RELATIONAL_HPP_
#include "relation.hpp"

#ifndef INCLUDE_BRUTEFORCE_CPP_DEPENDENCY
    #error "not intended for direct inclusion, include bruteforce.hpp instead"
#endif

typedef relation::matrix8 matrix8;
typedef relation::matrix16 matrix16;
typedef relation::adjacency_bitset_list<std::uint32_t> adjacency_bitset_list_32;
typedef relation::adjacency_bitset_list<std::uint64_t> adjacency_bitset_list_64;
typedef relation::packed_edges_list<std::uint16_t, 4, false> packed_edges_list_16_4;
typedef relation::edge_list<std::uint8_t> edge_list_8;
typedef relation::edge_list<std::uint16_t> edge_list_16;
typedef relation::edge_list<std::uint32_t> edge_list_32;


/**
 * The following types are intended to simplify the plugin nature of the
 * bruteforce_compute function with respect to internally-used relation
 * representations: via these structures we can provide a generic execute-
 * operation infrastructure, without running into problems with function
 * overloads when specifying the operations performed by this generic
 * infrastructure. As a consequence we can specialize based on the needs of the
 * specific burteforce task at hand: operations are either performed on
 * relations (when determining path separation) or on pairs of relations (when
 * determining boolean separation). In this way, the following structures are
 * comparable to the general std::less<> structs used for Heterogeneous lookups.
 * Each structure provides a single construct- or perform-function (that works
 * with and without buffers).
 */

/* The following types provide a type name for each possible construction
 * operation needed by the bruteforce_compute function. */

struct op_identity {
    template<class RelationType, class... BufferType>
    static RelationType construct(const std::size_t num_nodes, BufferType&... buffer) {
        return RelationType::identity(num_nodes, buffer...);
    }
};

struct op_diversity {
    template<class RelationType, class... BufferType>
    static RelationType construct(const std::size_t num_nodes, BufferType&... buffer) {
        return RelationType::diversity(num_nodes, buffer...);
    }
};

struct op_relation {
    template<class RelationType, class... BufferType>
    static RelationType construct(const raw_relation& relation, BufferType&... buffer) {
        return RelationType::relation(relation, buffer...);
    }
};


/* The following types provide a type name for each possible operation performed
 * by the bruteforce_compute function. */

struct op_converse {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& relation, BufferType&... buffer) {
        return RelationType::relation_converse(relation, buffer...);
    }
};

struct op_project_1 {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& relation, BufferType&... buffer) {
        return RelationType::relation_project_1(relation, buffer...);
    }
};

struct op_project_2 {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& relation, BufferType&... buffer) {
        return RelationType::relation_project_2(relation, buffer...);
    }
};

struct op_union {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& lhs, const RelationType& rhs, BufferType&... buffer) {
        return RelationType::relation_union(lhs, rhs, buffer...);
    }
};

struct op_compose {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& lhs, const RelationType& rhs, BufferType&... buffer) {
        return RelationType::relation_compose(lhs, rhs, buffer...);
    }
};

struct op_difference {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& lhs, const RelationType& rhs, BufferType&... buffer) {
        return RelationType::relation_difference(lhs, rhs, buffer...);
    }
};

struct op_intersect {
    template<class RelationType, class... BufferType>
    static RelationType perform(const RelationType& lhs, const RelationType& rhs, BufferType&... buffer) {
        return RelationType::relation_intersect(lhs, rhs, buffer...);
    }
};

#endif