/**
 * Copyright (c) 2014, 2015, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <cctype>
#include <map>
#include <stdexcept>
#include "bruteforce.hpp"
#include "string_parser.hpp"

/* Return true if ch is not a newline character. */
bool not_newline(char ch) {
    return ch != '\n' && ch != '\r';
}

/* Read an alphanumerical literal value, and return the literal on success. */
optional<std::string> read_literal(string_reader& reader) {
    return parser::read_function(reader, std::isalnum);
}

/* Skip whitespace and comments in the reader. Comments start with '#', and end
 * at the end of the line. */
void read_skippable(string_reader& reader) {
    while (!reader.end()) {
        auto ch = reader.peek_char();
        if (ch == '#') {
            while (!reader.end() && not_newline(reader.peek_char())) {
                reader.next();
            }
        }
        else if (std::isspace(ch)) {
            reader.next();
        }
        else {
            return;
        }
    }
}


/* Helper function for mapping strings to identifiers: map the provided name to
 * a single identifier such that name_map[name] provides this identifier and
 * names[identifier] provides the name. Return the already existing or newly
 * assigned identifier. */
template<class MapType, class ListType>
typename MapType::mapped_type name_to_id(const std::string& name, MapType& name_map, ListType& names) {
    typedef typename MapType::mapped_type uint_type;
    if (std::numeric_limits<uint_type>::max() == names.size()) {
        throw std::overflow_error("to many nodes, ran out of node identifiers");
    }

    auto result = name_map.emplace(name, uint_type(names.size()));
    if (result.second) {
        names.emplace_back(name);
    }

    return result.first->second;
}

/* Read a language configuration from the specified parser. */
template<class ParserType>
std::vector<configuration> read_languages(ParserType& parser) {
    std::vector<configuration> configurations;
    if (!parser.parse("languages", "{")) {
        parser.throw_error("unexpected input, expected languages declaration");
    }

    const std::map<std::string, bool configuration::*> names_to_field {
            { "diversity", &configuration::diversity },
            { "converse", &configuration::converse },
            { "project", &configuration::project },
            { "coproject", &configuration::coproject },
            { "intersect", &configuration::intersect },
            { "difference", &configuration::difference }
    };

    while (!parser.parse("}")) {
        configurations.emplace_back();
        configuration& new_configuration = configurations.back();

        while (!parser.parse(";")) {
            std::string name;
            bool parsed = parser.parse(parser::capture(name, read_literal));
            auto value = names_to_field.find(name);

            if (!parsed || value == names_to_field.end()) {
                parser.throw_error("unexpected input, operation expected");
            }

            bool& field = new_configuration.*(value->second);
            if (field) {
                parser.throw_error("duplicate operation");
            }
            field = true;
        }
    }
    return configurations;
}


/* Read a graph from the specified parser. */
template<class ParserType>
graph read_graph(ParserType& parser) {
    std::string graph_name;
    if (!parser.parse("graph", parser::capture(graph_name, read_literal),  "{")) {
        parser.throw_error("unexpected input, expected graph declaration");
    }

    std::map<std::string, node_id_type> name_node_map;
    std::vector<std::string> node_names;
    std::map<std::string, relation_id_type> name_relation_map;
    std::vector<std::string> relation_names;

    std::vector<raw_relation> relations;

    while (!parser.parse("}")) {
        std::string from, label, to;
        if (!parser.parse("(", parser::capture(from, read_literal), ",",
                               parser::capture(label, read_literal), ",", 
                               parser::capture(to, read_literal), ")")) {
                parser.throw_error("unexpected input, edge expected");
        }

        auto relation_id = name_to_id(label, name_relation_map, relation_names);
        auto from_id = name_to_id(from, name_node_map, node_names);
        auto to_id = name_to_id(to, name_node_map, node_names);

        if (relation_id == relations.size()) {
            relations.emplace_back();
        }

        relations[relation_id].emplace_back(from_id, to_id);
    }

    for (auto it = relations.begin(); it != relations.end(); ++it) {
        auto& relation = *it;
        std::sort(relation.begin(), relation.end());
        auto uit = std::unique(relation.begin(), relation.end());
        if (uit != relation.end()) {
            parser.throw_error("relation " + relation_names[it - relations.begin()] + " has duplicate edges");
        }
    }

    return graph { graph_name, node_names, relation_names, relations };
}

/* Read a single script from the specified parser. */
template<class ParserType>
std::unique_ptr<script> read_script(ParserType& parser) {
    std::unique_ptr<script> result(nullptr);
    std::string mode, identifier;
    if (!parser.parse(parser::capture(mode, read_literal),
                        parser::capture(identifier, read_literal), "{")) {
        parser.throw_error("invalid begin of script (path/bool)");
    }

    if (mode == "path") {
        auto configuration = read_languages(parser);
        auto graph = read_graph(parser);
        result = std::make_unique<path_script>(identifier, configuration, graph);
    }
    else if (mode == "boolean") {
        auto configuration = read_languages(parser);
        auto first_graph = read_graph(parser);
        auto second_graph = read_graph(parser);
        result = std::make_unique<boolean_script>(identifier, configuration,
                                                  first_graph, second_graph);
    }
    else {
        parser.throw_error("unknown script mode " + mode + " (path/bool expected)");
    }

    if (!parser.parse("}")) {
        parser.throw_error("expected closing curly bracket `}' (end-of-script)");  
    }

    return result;
}


std::vector<std::unique_ptr<script>>
read_scripts(const std::string& source) {
    std::vector<std::unique_ptr<script>> result;
    auto parser = parser::create(source, read_skippable);
    while (!parser.end()) {
        result.emplace_back(read_script(parser));
    }
    return result;
}