/**
 * Copyright (c) 2013, 2014, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_BRUTEFORCE_TREE_HPP_
#define INCLUDE_BRUTEFORCE_TREE_HPP_
#include <cstdint>
#include <cstddef>
#include "graph.hpp"

#ifndef INCLUDE_BRUTEFORCE_CPP_DEPENDENCY
    #error "not intended for direct inclusion, include bruteforce.hpp instead"
#endif

/* Bruteforce operation trees are used to describe the expression (as a syntax
 * tree) of the expression resulting in a specific query result. The trees have
 * four kinds of nodes: binary operation nodes (the relational composition,
 * union, intersection, and difference operations), unary operation nodes (the
 * projection and coprojection nodes), relation atom nodes (relations and
 * converse-relations), and constant nodes (identity and diversity). The binary
 * and unary nodes contain children, these children are represented as an index
 * (the brute force process maintains a vector of all generated entries, each
 * entry consists of the generated query results and the corresponding operation
 * node). */

/* Type of operations. */
enum class operation_type : std::uint8_t {
    identity,
    diversity,
    relation,
    converse_relation,
    project_1,
    project_2,
    coproject_1,
    coproject_2,
    relation_compose,
    relation_union,
    relation_intersect,
    relation_difference
};

/* Data fields for binary, unary and relational nodes. */
struct binary_operation {
    std::size_t lhs_id;
    std::size_t rhs_id;
};

struct unary_operation {
    std::size_t child_id;
};

struct relation_operation {
    std::size_t relation_id;
};

/* The bruteforce operation tree nodes .*/
class bruteforce_tree {
public:
    /* Construct constant tree nodes. */
    bruteforce_tree(const operation_type operation) :
            operation(operation) { }

    /* Construct relation and conversed relation tree nodes. */
    bruteforce_tree(const std::size_t relation_id, bool converse) :
            operation(converse ? operation_type::converse_relation : operation_type::relation),
            relation{relation_id} { }

    /* Construct unary tree nodes. */
    bruteforce_tree(const operation_type operation, const std::size_t child_id) :
            operation(operation), unary{child_id} { }

    /* Construct binary tree nodes. */
    bruteforce_tree(const operation_type operation,
                    const std::size_t lhs_id,
                    const std::size_t rhs_id) :
            operation(operation), binary{lhs_id, rhs_id} { }


    /* Write the expression to the specified output stream, using the provided
     * vector of bruteforce entries as the lookup for child nodes (by their
     * position in the vector). The provied graph provides mapping between
     * relation ids and relation names. */
    template<class VectorType>
    void to_output(const VectorType& nodes,
                   const std::vector<std::string>& relation_names,
                   std::ostream& output) const {
        switch (operation) {
            case operation_type::identity:
                output << "id";
                return;

            case operation_type::diversity:
                output << "di";
                return;

            case operation_type::relation:
                output << '<' << relation_names[relation.relation_id] << '>';
                return;

            case operation_type::converse_relation:
                output << "~<" << relation_names[relation.relation_id] << '>';
                return;

            case operation_type::project_1:
                to_output_unary(nodes, relation_names, output, "project_1");
                return;

            case operation_type::project_2:
                to_output_unary(nodes, relation_names, output, "project_2");
                return;

            case operation_type::coproject_1:
                to_output_unary(nodes, relation_names, output, "coproject_1");
                return;

            case operation_type::coproject_2:
                to_output_unary(nodes, relation_names, output, "coproject_2");
                return;

            case operation_type::relation_compose:
                to_output_binary(nodes, relation_names, output, "compose");
                return;

            case operation_type::relation_union:
                to_output_binary(nodes, relation_names, output, "union");
                return;

            case operation_type::relation_intersect:
                to_output_binary(nodes, relation_names, output, "intersect");
                return;

            case operation_type::relation_difference:
                to_output_binary(nodes, relation_names, output, "difference");
                return;
        }
    }

    /* Operation (type) of this tree. */
    operation_type operation;

    /* Data fields of the node. */
    union {
        const relation_operation relation;
        const unary_operation unary;
        const binary_operation binary;
    };

private:
    /* Output formatting for binary rules. */
    template<class VectorType>
    void to_output_binary(const VectorType& nodes, const std::vector<std::string>& relation_names,
                          std::ostream& output, const char operator_name[]) const {
        output << '(';
        nodes[binary.lhs_id]->second.to_output(nodes, relation_names, output);
        output << ' ' << operator_name << ' ';
        nodes[binary.rhs_id]->second.to_output(nodes, relation_names, output);
        output << ')';
    }

    /* Output formatting for unary rules. */
    template<class VectorType>
    void to_output_unary(const VectorType& nodes, const std::vector<std::string>& relation_names,
                         std::ostream& output, const char operator_name[]) const {
        output << operator_name << '[';
        nodes[unary.child_id]->second.to_output(nodes, relation_names, output);
        output << ']';
    }
};

#endif