/**
 * Copyright (c) 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_UTILITY_HPP_
#define INCLUDE_UTILITY_HPP_
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>

/* Hold a value, optionally (if no value is present, then retrieving the value
 * will return a default-constructed value). */
template<class ValueType>
class optional {
public:
    typedef ValueType value_type;

    /* Construct an optional that does not contain a value. */
    optional() : value(), valid(false) { }

    /* Construct an optional that does contain a value (copy the value). */
    optional(const value_type& value) : value(value), valid(true) { }

    /* Construct an optional that does contain a value (move the value). */
    optional(value_type&& value) : value(std::move(value)), valid(true) { }

    /* Copy-constructor. */
    optional(const optional<value_type>& other) : value(other.value), valid(other.valid) { }

    /* Move-constructor. */
    optional(optional<value_type>&& other) : value(std::move(other.value)), valid(other.valid) { }


    /* Return true if this optional holds a value. */
    bool has_value() const {
        return valid;
    }


    /* Return the held value (default-constructed if !has_value()). */
    const value_type& get_value() const {
        return value;
    }

    /* Return the held value (default-constructed if !has_value()). */
    value_type& get_value() {
        return value;
    }

private:
    value_type value;
    bool valid;
};

#endif