/**
 * Copyright (c) 2013, 2014, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_RELATION_HPP_
#define INCLUDE_RELATION_HPP_
#include <array>
#include <algorithm>
#include <cstddef>
#include <cstdint>
#include <iterator>
#include "array.hpp"
#include "bit_iterator.hpp"
#include "graph.hpp"
#include "overflow.hpp"

/* The matrix implementations heavily relies on the x86 memory layout of data
 * types, on SSE, AVX, and ther standard SSE and AVX Intrinsics. */
#include "immintrin.h"

/*
 * We have several possible implementations for binary node relations that each
 * provide a trade of between performance, memory usage, and supported graph
 * size (in the number of nodes). Each relation implementation provides the same
 * relational operators; which are described for the (non-implemented)
 * relation_interface class.
 *
 * We remark that, globally speaking, we make a distinction between
 * implementations that perform every operation fully in-place and those
 * implementations that are not in-place and, for performance reasons, rely on
 * an externally provided buffer of an appropriate size. If X is a class
 * providing a relation implementation, then X::buffer_type specifies the type
 * of the buffer, which is relation_inplace_tag if the implementation is
 * in-place.
 *
 * The documented relation_interface class provides the interface for the
 * in-place implementations. Implementations that rely on buffers differ in the
 * following way:
 *
 *   - they must provide a static method to create a buffer of an appropriate
 *     size for any relational operation operating on relation(s) of nodes in a
 *     graph with at most max_num_nodes nodes:
 *
 *         static buffer_type initial_buffer(const std::size_t max_num_nodes);
 *
 *   - each relational operation has an additional argument buffer_type& buffer.
 *     Hence, for example, instead of
 *
 *         static relation_type relation(const raw_relation& relation);
 *
 *     we have
 *
 *         static relation_type relation(const raw_relation& relation,
 *                                       buffer_type& buffer);
 *
 * Buffers can safely be shared across sequentially executed relational
 * operations. In a multithreaded environment, each thread requires its own
 * buffer, which is why the buffer is explicitly provided as input.
 */

 namespace relation {
    /* Tag used to indicate that all relational operators of the relation
     * implementation works in-place. */
    class relation_inplace_tag { };

    /* General interface provided by all relational implementations. Observe:
     * this is not a base class (we depend on template metaprogramming instead
     * of object-oriented inheritance to support various relation
     * implementations). */
    class relation_interface {
    public:
        /* Type of this relation implementation. */
        typedef relation_interface relation_type;

        /* Type of the buffer used by this relational implementations. */
        typedef relation_inplace_tag buffer_type;

        /* Upper bound on the number of nodes supported by this relation
         * implementation. */
        static constexpr std::size_t max_num_nodes = 0;


        /* Construct an empty relation. */
        relation_interface();

        /* Copy-construct a relation. */
        relation_interface(const relation_type& other);

        /* Move-construct a relation. */
        relation_interface(relation_type&& other);


        /* Create a buffer large enough to accommodate all additional memory
         * requirements of all relational operators on relations build over the
         * specified maximum number of elements. This method is only present if
         * the implementation is not in place (the buffer type is not
         * relation_inplace_tag). */
        /*
           static buffer_type initial_buffer(const std::size_t max_num_nodes);
         */


        /* Provide a strict ordering on relations. The ordering provided by the
         * relation implementations is not guaranteed to be consistent: each
         * implementation can use a different ordering due to performance
         * reasons; the only guarantee is that the provided ordering is strict
         * with regards to all relations represented by the same relation
         * implementation. */
        bool operator<(const relation_type& other) const;

        /* Return true if the relation is empty. */
        bool is_empty() const;


        /* Make a relation_type-copy of the provided raw relation. */
        static relation_type relation(const raw_relation& relation /*, buffer_type& buffer */);

        /* Return the raw_relation representation of this relation. */
        raw_relation to_relation() const;


        /* Return the identity relation for a graph with n nodes, which is
         * defined as { (i, i) | 0 <= i < n }. */
        static relation_type identity(const std::size_t n /*, buffer_type& buffer */);

        /* Return the diversity relation for a graph with n nodes, which is
         * defined as { (i, j) | 0 <= i < n, 0 <= j < n, i != j }. */
        static relation_type diversity(const std::size_t n /*, buffer_type& buffer */);


        /* Return the converse (the transpose) of the specified relation, which
         * is defined as { (m, n) | (n, m) in other }. */
        static relation_type relation_converse(const relation_type& relation /*, buffer_type& buffer */);

        /* Return the first projection of the specified relation,  which is
         * defined as { (n, n) | exists m with (n, m) in other }. */
        static relation_type relation_project_1(const relation_type& relation /*, buffer_type& buffer */);

        /* Return the second projection of the specified relation,  which is
         * defined as { (m, m) | exists n with (n, m) in other }. */
        static relation_type relation_project_2(const relation_type& relation /*, buffer_type& buffer */);


        /* Return the relational composition of the two relations, which is
         * defined as { (n, m) | exists v with (n, v) in lhs and (v, m) in rhs
         * }.*/
        static relation_type relation_compose(const relation_type& lhs, const relation_type& rhs /*, buffer_type& buffer */);

        /* Return the relational union of the two relations, which is defined as
         * { (n, m) | (n, m) in lhs or (n, m) in rhs }. */
        static relation_type relation_union(const relation_type& lhs, const relation_type& rhs /*, buffer_type& buffer */);

        /* Return the relational intersection of the two relations, which is
         * defined as { (n, m) | (n, m) in lhs,  (n, m) in rhs }. */
        static relation_type relation_intersect(const relation_type& lhs, const relation_type& rhs /*, buffer_type& buffer */);

        /* Return the relational difference of the two relations, which is
         * defined as { (n, m) | (n, m) in lhs, (n, m) not in rhs }. */
        static relation_type relation_difference(const relation_type& lhs, const relation_type& rhs /*, buffer_type& buffer */);
    };

    /* Forward declarations of each relation implementation. */
    class matrix8;
    class matrix16;
    template<class NodeType> class edge_list;
    template<class NodeType> class adjacency_bitset_list;
    template<class NodeType,
             const std::size_t PackingFactor,
             bool PairRepresentation> class packed_edges_list;

    /* Forward declarations of the details used by the implementations. */
    namespace details   {
        /* Used by matrix8 and matrix16. */
        union field_int;
        union field_64;
        union field_128;
        union field_256;

        /* Used by edge_list. */
        template<class UIntType> class internal_edge_traits;
        template<class NodeType> class internal_edge;

        /* Used by adjacency_bitset_list. */
        template<class NodeType> class adjacency_bitset_node_traits;
        template<class NodeType> class adjacency_bitset_node;

        /* Used by packed_edge_list. */
        template<std::size_t PackingFactor> struct packed_edges_traits;
        template<class NodeType, std::size_t PackingFactor> class packed_edges_plain;
        template<class NodeType, std::size_t PackingFactor> class packed_edges_pair;

        /* Algorithms used by adjacency_bitset_list and packed_edge_list. */
        template<class KeyType, class BitsetType> class key_bitset_pair;

        template<class InIt, class OutIt>
        OutIt key_bitset_union(InIt lhs_it, InIt lhs_end,
                               InIt rhs_it, InIt rhs_end, OutIt result);
        template<class InIt, class OutIt>
        OutIt key_bitset_intersection(InIt lhs_it, InIt lhs_end,
                                      InIt rhs_it, InIt rhs_end, OutIt result);
        template<class InIt, class OutIt>
        OutIt key_bitset_difference(InIt lhs_it, InIt lhs_end,
                                    InIt rhs_it, InIt rhs_end, OutIt result);

    }
}

#include "relation_matrix.hpp"
#include "relation_edge_list.hpp"
#include "relation_key_bitset.hpp"

#endif
