/**
 * Copyright (c) 2013, 2014, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* Structure providing the unsigned integer type used to store edges in (will
 * provide exactly enough room for two nodes). */
template<> struct relation::details::internal_edge_traits<std::uint8_t> {
    typedef std::uint16_t edge_type;
};

template<> struct relation::details::internal_edge_traits<std::uint16_t> {
    typedef std::uint32_t edge_type;
};

template<> struct relation::details::internal_edge_traits<std::uint32_t> {
    typedef std::uint64_t edge_type;
};

/* An edge is a pair of nodes. */
template<class NodeType>
class relation::details::internal_edge {
public:
    /* Type of the nodes. */
    typedef NodeType node_type;

    /* Type big enough to fit exactly two nodes (edge-representation). */
    typedef typename internal_edge_traits<node_type>::edge_type edge_type;

    /* Construct an empty edge. */
    internal_edge() : value(0) { }

    /* Construct the edge (from, to). */
    internal_edge(const node_type& from, const node_type& to) :
            to(to), from(from) {}

    /* Return the from-node. */
    node_type node_from() const {
        return from;
    }

    /* Return the to-node. */
    node_type node_to() const {
        return to;
    }


    /* Return true if this edge and other are equivalent. */
    bool operator==(const internal_edge<node_type>& other) const {
        return value == other.value;
    }

    /* Return true if this edge comes lexicographically before other. */
    bool operator<(const internal_edge<node_type>& other) const {
        return value < other.value;
    }

    /* Return true if this edge comes lexicographically before any outgoing edge
     * of node. */
    bool operator<(const node_type& node) const {
        return from < node;
    }

private:
    /* We use the union of a logical representation of (from, to)-nodes and the
     * physical representation. We use the physical representation to optimize
     * comparison operations between edges. */
    union {
        struct {
            node_type to;
            node_type from;
        };
        edge_type value;
    };
};


/* Traditional implementation of a relation as a list of edge-pairs; for
 * performance reasons, the internal representation is guaranteed to be sorted
 * (lexicographically on (from, to)-edge pairs). This implementation relies on a
 * buffer whose size has a quadratic upper bounded on the number of nodes. In
 * this implementation, a node can be an unsigned 8bit, 16bit, or 32 bit integer
 * (std::uint8_t, std::uint16_t, or std::uint32_t). */
template<class NodeType>
class relation::edge_list {
private:
    typedef NodeType node_type;
    typedef details::internal_edge<node_type> internal_edge_type;
    typedef array<internal_edge_type> internal_edge_list;

public:
    typedef edge_list<node_type> relation_type;
    typedef internal_edge_list buffer_type;
    static constexpr std::size_t max_num_nodes = std::numeric_limits<node_type>::max();

    edge_list() : edges() { }

    edge_list(const relation_type& other) : edges(other.edges) { }

    edge_list(relation_type&& other) : edges(std::move(other.edges)) { }

    static buffer_type initial_buffer(const std::size_t max_num_nodes) {
        auto value = overflow_multiply<std::size_t>(2, max_num_nodes, max_num_nodes);
        if (value.second) {
            throw std::bad_alloc();
        }
        return buffer_type(value.first);
    }

    bool operator<(const relation_type& other) const {
        return edges.size() < other.edges.size() ||
                (edges.size() == other.edges.size() && edges < other.edges);
    }

    bool is_empty() const {
        return edges.size() == 0;
    }

    static relation_type relation(const raw_relation& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto& pair : relation) {
            *it = internal_edge_type(pair.from, pair.to);
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    raw_relation to_relation() const {
        raw_relation result;
        for (auto edge : edges) {
            result.emplace_back(static_cast<node_id_type>(edge.node_from()),
                                static_cast<node_id_type>(edge.node_to()));
        }
        return result;
    }

    static relation_type identity(const std::size_t n, buffer_type& buffer) {
        auto it = buffer.begin();
        for (node_type node = 0; node < n; ++node) {
            *it = internal_edge_type(node, node);
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type diversity(const std::size_t n, buffer_type& buffer) {
        auto it = buffer.begin();
        for (node_type node_from = 0; node_from < n; ++node_from) {
            for (node_type node_to = 0; node_to < node_from; ++node_to) {
                *it = internal_edge_type(node_from, node_to);
                ++it;
            }
            for (node_type node_to = node_from + 1; node_to < n; ++node_to) {
                *it = internal_edge_type(node_from, node_to);
                ++it;
            }
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_converse(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto edge : relation.edges) {
            *it = internal_edge_type(edge.node_to(), edge.node_from());
            ++it;
        }
        std::sort(buffer.begin(), it);
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_project_1(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        auto rit = relation.edges.begin(), rend = relation.edges.end();
        if (rit != rend) {
            *it = internal_edge_type(rit->node_from(), rit->node_from());
            ++rit;
            while (rit != rend) {
                if (it->node_from() != rit->node_from()) {
                    ++it;
                    *it = internal_edge_type(rit->node_from(), rit->node_from());
                }
                ++rit;
            }
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_project_2(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto edge : relation.edges) {
            *it = internal_edge_type(edge.node_to(), edge.node_to());
            ++it;
        }
        std::sort(buffer.begin(), it);
        return relation_type(buffer.begin(), std::unique(buffer.begin(), it));
    }

    static relation_type relation_compose(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto it = buffer.begin();
        auto lhs_it = lhs.edges.begin();
        while (lhs_it != lhs.edges.end()) {
            auto it_begin = it;
            auto node_from = lhs_it->node_from();
            auto rhs_it = rhs.edges.begin();
            do {
                auto node = lhs_it->node_to();
                rhs_it = std::lower_bound(rhs_it, rhs.edges.end(), node);
                while (rhs_it != rhs.edges.end() && rhs_it->node_from() == node) {
                    *it = internal_edge_type(node_from, rhs_it->node_to());
                    ++it;
                    ++rhs_it;
                }
                ++lhs_it;
            } while (lhs_it != lhs.edges.end() && lhs_it->node_from() == node_from);
            std::sort(it_begin, it);
            it = std::unique(it_begin, it);
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_union(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = std::set_union(lhs.edges.begin(), lhs.edges.end(),
                                  rhs.edges.begin(), rhs.edges.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

    static relation_type relation_intersect(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = std::set_intersection(lhs.edges.begin(), lhs.edges.end(),
                                         rhs.edges.begin(), rhs.edges.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

    static relation_type relation_difference(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = std::set_difference(lhs.edges.begin(), lhs.edges.end(),
                                       rhs.edges.begin(), rhs.edges.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

private:
    internal_edge_list edges;

    /* Construct an edge list from a sorted list of internal edges. */
    template<class InputIterator>
    edge_list(InputIterator first, InputIterator last) : edges(first, last) { }
};