/**
 * Copyright (c) 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Both the adjacency_bitset_list and the packed_edge_list represent packed edge
 * lists in which each list element is a (node, children)-pair. These elements
 * are represented by (key, bitset)-pairs. In the adjacency_bitset_list each
 * list element represents all outgoing edges of a single node, whereas the
 * packed_edge_list represents a smaller set of outgoing edges (resulting in
 * several list elements describing outgoing edges of a single node).
 *
 * For these relation implementations, the union, intersections, and differences
 * relation operations operate on lists of elements in which each physical
 * element represents several logical elements. To take this logical
 * representation into account, we expect the individual list elements to
 * implement the following functionality:
 */
template<class KeyType, class BitsetType>
class relation::details::key_bitset_pair {
public:
    /* The type of the keys and bitsets used in the pair. */
    typedef KeyType key_type;
    typedef BitsetType bitset_type;
    typedef key_bitset_pair<key_type, bitset_type> pair_type;

    /* Construct a (key, bitset)-pair. */
    key_bitset_pair(const key_type key, const bitset_type btifield);

    /* Return the key for the provided (key, bitset)-pair. */
    static key_type get_key(const pair_type& pair);

    /* Return the key for the provided (key, bitset)-pair. */
    static key_type get_bitset(const pair_type& pair);
};

/* Construct the union of the two sorted lists of (key, bitset)-pairs such that
 * the bitsets of pairs with equal keys are merged and (key, bitset)-pairs for
 * which a key only occurs in one of the two sorted lists are kept. Return the
 * one-past-last iterator to the position in result after writing the union to
 * the result iterator. */
template<class InIt, class OutIt>
OutIt relation::details::key_bitset_union(InIt lhs_it, InIt lhs_end,
                                          InIt rhs_it, InIt rhs_end, OutIt result) {
    typedef std::iterator_traits<OutIt>::value_type value_type;
    while (lhs_it != lhs_end && rhs_it != rhs_end) {
        auto lhs_key = value_type::get_key(*lhs_it);
        auto rhs_key = value_type::get_key(*rhs_it);
	    if (lhs_key < rhs_key) {
		    *result = *lhs_it;
            ++lhs_it;
	    }
        else if (rhs_key < lhs_key) {
		    *result = *rhs_it;
            ++rhs_it;
        }
        else {
            auto field = value_type::get_bitset(*lhs_it) | value_type::get_bitset(*rhs_it);
		    *result = value_type(lhs_key, field);
            ++lhs_it;
            ++rhs_it;
	    }
        ++result;
    }
    result = std::copy(lhs_it, lhs_end, result);
	return std::copy(rhs_it, rhs_end, result);
}

/* Construct the intersection of the two sorted lists of (key, bitset)-pairs
 * such that a pair is only kept when the key occurs in both sorted lists and
 * the intersection of the bitsets is non-empty. Return the one-past-last
 * iterator to the position in result after writing the intersection to the
 * result iterator. */
template<class InIt, class OutIt>
OutIt relation::details::key_bitset_intersection(InIt lhs_it, InIt lhs_end,
                                                 InIt rhs_it, InIt rhs_end, OutIt result) {
    typedef std::iterator_traits<OutIt>::value_type value_type;
    while (lhs_it != lhs_end && rhs_it != rhs_end) {
        auto lhs_key = value_type::get_key(*lhs_it);
        auto rhs_key = value_type::get_key(*rhs_it);
	    if (lhs_key < rhs_key) {
            ++lhs_it;
	    }
        else if (rhs_key < lhs_key) {
            ++rhs_it;
        }
        else {
            auto field = value_type::get_bitset(*lhs_it) & value_type::get_bitset(*rhs_it);
            if (field != 0) {
		        *result = value_type(lhs_key, field);
                ++result;
            }
            ++lhs_it;
            ++rhs_it;
	    }
    }
	return result;
}

/* Construct the difference of the two sorted lists of (key, bitset)-pairs such
 * that a pair is only kept when the key occurs in the first list and not in the
 * second, or, on equal keys, when the difference of the left bitset and the
 * right bitset is non-empty.  Return the one-past-last iterator to the position
 * in result after writing the difference to the result iterator. */
template<class InIt, class OutIt>
OutIt relation::details::key_bitset_difference(InIt lhs_it, InIt lhs_end,
                                               InIt rhs_it, InIt rhs_end, OutIt result) {
    typedef std::iterator_traits<OutIt>::value_type value_type;
    while (lhs_it != lhs_end && rhs_it != rhs_end) {
        auto lhs_key = value_type::get_key(*lhs_it);
        auto rhs_key = value_type::get_key(*rhs_it);
	    if (lhs_key < rhs_key) {
		    *result = *lhs_it;
            ++lhs_it;
            ++result;
	    }
        else if (rhs_key < lhs_key) {
            ++rhs_it;
        }
        else {
            auto field = value_type::get_bitset(*lhs_it) & ~value_type::get_bitset(*rhs_it);
            if (field != 0) {
		        *result = value_type(lhs_key, field);
                ++result;
            }
            ++lhs_it;
            ++rhs_it;
	    }
    }
    return std::copy(lhs_it, lhs_end, result);
}



/* In the adjacency_bitset_list each list entry is a (node, children)-pair and
 * this pair is encoded in an unsigned integer of a fixed width, which limits
 * the maximum number of nodes supported by this relation implementation. The
 * following structure provides the number of nodes that fit in the specified
 * unsigned integer type. */
template<>
struct relation::details::adjacency_bitset_node_traits<std::uint32_t> {
    static constexpr std::uint32_t max_num_nodes = 27;
};

template<>
struct relation::details::adjacency_bitset_node_traits<std::uint64_t> {
    static constexpr std::uint64_t max_num_nodes = 58;
};

/* Single-node entries used in the adjacency_bitset_list. Each entry is of the
 * form (node, children), in which children is represented by a bitset in which
 * the j-th bit of the bitset is set if there is an outgoing edge (node, j). */
template<class NodeType>
class relation::details::adjacency_bitset_node {
public:
    typedef NodeType node_type;
    typedef node_type key_type;
    typedef node_type bitset_type;
    typedef adjacency_bitset_node_traits<node_type> traits;
    typedef adjacency_bitset_node<node_type> pair_type;

    static constexpr node_type one = 1;
    static constexpr node_type max_num_nodes = traits::max_num_nodes;

    /* Construct an empty (node, children)-pair. */
    adjacency_bitset_node() : data(0) { }

    /* Construct an (node, children)-pair without any children. */
    adjacency_bitset_node(const key_type node_from) :
            data(node_from << key_shift) { }

    /* Construct a (node, children)-pair from the specified key (node) and 
     * bitset (children). */
    adjacency_bitset_node(const key_type key, const bitset_type bitset) :
            data((key << key_shift) | bitset) { }

    /* Return the (node, childrenn0-pair representing the single edge
     * (node_from, node_to). */
    static pair_type create_edge(const node_type node_from, const node_type node_to) {
        return adjacency_bitset_node(node_from, one << node_to);
    }


    /* Get the from-node. */
    node_type node_from() const {
        return data >> key_shift;
    }

    /* Get the bitset representing the children. */
    bitset_type nodes_to_field() const {
        return data & to_field_mask;
    }

    /* Add node to the list of children. */
    void add_node_to(const node_type node_to) {
        data |= (one << node_to);
    }


    /* Provide a strict ordering on (node, children)-pairs. */
    bool operator<(const pair_type& other) const {
        return data < other.data;
    }

    /* Return true if the other (node, children)-pairs is equivalent to this
     * (node, children)-pair. */
    bool operator==(const pair_type& other) const {
        return data == other.data;
    }


    /* key_bitset_pair interface get_key-function, as used by the
     * key_bitset-algorithms. */
    static node_type get_key(const pair_type& entry) {
        return entry.node_from();
    }

    /* key_bitset_pair interface get_bitset-function, as used by the
     * key_bitset-algorithms. */
    static node_type get_bitset(const pair_type& entry) {
        return entry.nodes_to_field();
    }

    
private:
    /* Parameters used for the bitwise implementation: in the (node,
     * children)-pair, the children-bitset are the first max_num_nodes-bits, the
     * remaining bits are used to store the from-node. The key_shift provides
     * the number of bits the from-node is shifted and the to_field_mask is a
     * bitwise mask used to filter out a children-bitset from the underlying
     * unsigned integer. */
    static constexpr node_type key_shift = max_num_nodes;
    static constexpr node_type to_field_mask = (one << key_shift) - 1;

    /* The (node, children)-pair. */
    node_type data;
};

/* Compact adjacency-list implementation of a relation: we maintain an list of
 * nodes-entries, in which each node-entry is an unsigned integer encoding pair
 * (node n, children of n). The list of node entries is ordered on node n. This
 * implementation relies on a buffer whose size has a linear upper bounded on
 * the number of nodes. */
template<class NodeType>
class relation::adjacency_bitset_list {
private:
    typedef NodeType node_type;
    typedef details::adjacency_bitset_node<node_type> node_entry_type;
    typedef array<node_entry_type> internal_node_list;

public:
    typedef adjacency_bitset_list<node_type> relation_type;
    typedef internal_node_list buffer_type;
    static constexpr std::size_t max_num_nodes = node_entry_type::max_num_nodes;

    adjacency_bitset_list() : nodes() { }

    adjacency_bitset_list(const relation_type& other) : nodes(other.nodes) { }

    adjacency_bitset_list(relation_type&& other) : nodes(std::move(other.nodes)) { }

    static buffer_type initial_buffer(const std::size_t max_num_nodes) {
        return buffer_type(max_num_nodes);
    }

    bool operator<(const relation_type& other) const {
        return nodes.size() < other.nodes.size() ||
                (nodes.size() == other.nodes.size() && nodes < other.nodes);
    }

    bool is_empty() const {
        return nodes.size() == 0;
    }

    static relation_type relation(const raw_relation& relation, buffer_type& buffer) {
        if (relation.empty()) {
            return relation_type();
        }

        auto r_it = relation.begin(), r_end = relation.end();
        auto it = buffer.begin();
        *it = node_entry_type::create_edge(r_it->from, r_it->to);
        ++r_it;

        while (r_it != r_end) {
            if (r_it->from == it->node_from()) {
                it->add_node_to(r_it->to);
            }
            else {
                ++it;
                *it = node_entry_type::create_edge(r_it->from, r_it->to);
            }
            ++r_it;
        }
        ++it;
        return relation_type(buffer.begin(), it);
    }

    raw_relation to_relation() const {
        raw_relation result;
        for (auto node_entry : nodes) {
            auto node_from = node_entry.node_from();
            for (auto node_to : make_bit_iterator(node_entry.nodes_to_field())) {
                result.emplace_back(static_cast<node_id_type>(node_from),
                                    static_cast<node_id_type>(node_to));
            }
        }
        return result;
    }

    static relation_type identity(const std::size_t n, buffer_type& buffer) {
        auto it = buffer.begin();
        for (node_type node = 0; node < n; ++node) {
            *it = node_entry_type::create_edge(node, node);
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type diversity(const std::size_t n, buffer_type& buffer) {
        auto it = buffer.begin();
        node_type all_children = (node_entry_type::one << n) - 1;
        for (node_type node = 0; node < n; ++node) {
            node_type exclude_this_mask = ~(node_entry_type::one << node);
            node_type children_field = exclude_this_mask & all_children;
            *it = node_entry_type(node, children_field);
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_converse(const relation_type& relation, buffer_type& buffer) {
        node_type children_field = 0;
        for (auto& node : relation.nodes) {
            children_field |= node.nodes_to_field();
        }

        for (auto node : make_bit_iterator(children_field)) {
            buffer[node] = node_entry_type(node);
        }

        for (auto& node : relation.nodes) {
            auto node_from = node.node_from();
            for (auto node_to : make_bit_iterator(node.nodes_to_field())) {
                buffer[node_to].add_node_to(node_from);
            }
        }

        auto it = buffer.begin();
        for (auto node : make_bit_iterator(children_field)) {
            *it = buffer[node];
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_project_1(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto& node : relation.nodes) {
            auto node_value = node.node_from();
            *it = node_entry_type::create_edge(node_value, node_value);
            ++it;
        }
        return relation_type(buffer.begin(), std::unique(buffer.begin(), it));
    }

    static relation_type relation_project_2(const relation_type& relation, buffer_type& buffer) {
        node_type children_field = 0;
        for (auto& node : relation.nodes) {
            children_field |= node.nodes_to_field();
        }

        auto it = buffer.begin();
        for (auto node : make_bit_iterator(children_field)) {
            *it = node_entry_type::create_edge(node, node);
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_compose(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto& node : lhs.nodes) {
            auto node_from = node.node_from();
            auto rhs_it = rhs.nodes.begin();

            node_type to_set = 0;
            for (auto node_to : make_bit_iterator(node.nodes_to_field())) {
                rhs_it = std::lower_bound(rhs_it, rhs.nodes.end(), node_entry_type(node_to));
                if (rhs_it != rhs.nodes.end() && rhs_it->node_from() == node_to) {
                    to_set |= rhs_it->nodes_to_field();
                }
            }
            if (to_set != 0) {
                *it = node_entry_type(node_from, to_set);
                ++it;
            }
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_union(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = details::key_bitset_union(lhs.nodes.begin(), lhs.nodes.end(),
                                             rhs.nodes.begin(), rhs.nodes.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

    static relation_type relation_intersect(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = details::key_bitset_intersection(lhs.nodes.begin(), lhs.nodes.end(),
                                                    rhs.nodes.begin(), rhs.nodes.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

    static relation_type relation_difference(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = details::key_bitset_difference(lhs.nodes.begin(), lhs.nodes.end(),
                                                  rhs.nodes.begin(), rhs.nodes.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

private:
    internal_node_list nodes;

    /* Construct an adjacency list from a sorted list of internal nodes. */
    template<class InputIterator>
    adjacency_bitset_list(InputIterator first, InputIterator last) : nodes(first, last) { }
};



/* In the packed_edge_list each list entry is a triple (node n , base b,
 * bitfield f) representing the set of edges { (n, b + y } | the y-th bit is set
 * in f }. We provide two encodings of such a triplet: one in which an entire
 * triplet is encoded in a single unsigned integer of a fixed width (the plain
 * representation), and one in which the node and base are encoded in a single
 * unsigned integer of a fixed width and the bitfield is encoded in a second
 * unsigned integer of a fixed width (the pair representation). We call the size
 * of the bitset the packing factor (as it determines how many edges can be
 * represented by a single list element). If we use 2k bits to represent the
 * pair (n, b) with a packing factor of 4^i, then each node identifier must be a
 * k+i-bit integer (we use k+i bits for node n, k-i bits for base b, and at most
 * 4^i edges are stored in the bitfield).
 *
 * We support packing factors of 4, 16, and 64. A packing factor of 4 is only
 * supported in the plain representation, the packing factor 16 can be used in
 * plain and pair representations, and the packing factor 64 can only be used in
 * the pair representation. */
template<> struct relation::details::packed_edges_traits<4> {
    static constexpr std::size_t added_bits = 1;
    static constexpr std::uint8_t bitset_mask = 0xF;
};
template<> struct relation::details::packed_edges_traits<16> {
    typedef std::uint16_t bitset_type;
    static constexpr std::size_t added_bits = 2;
    static constexpr bitset_type bitset_mask = 0xFFFF;
};
template<> struct relation::details::packed_edges_traits<64> {
    typedef std::uint64_t bitset_type;
    static constexpr std::size_t added_bits = 3;
    static constexpr bitset_type bitset_mask = 0xFFFFFFFF;
};

/* Plain representation of packed edges. */
template<class NodeType, std::size_t PackingFactor>
class relation::details::packed_edges_plain {
public:
    typedef NodeType node_type;
    typedef node_type key_type;
    typedef node_type bitset_type;

    static constexpr node_type packing_factor = PackingFactor;
    typedef packed_edges_traits<packing_factor> traits;
    typedef packed_edges_plain<node_type, packing_factor> pair_type;

    static constexpr node_type one = 1;
    static constexpr node_type digits = std::numeric_limits<node_type>::digits;
    static constexpr node_type key_size = digits - packing_factor;
    static constexpr node_type node_size = (key_size / 2) + traits::added_bits;
    static constexpr node_type base_size = (key_size / 2) - traits::added_bits;
    static constexpr bitset_type bitset_mask = traits::bitset_mask;

    static constexpr node_type max_num_nodes = one << node_size;

    /* Construct an empty (node, base, bitset)-triplet. */
    packed_edges_plain() : data(0) { }

    /* Construct a (node, 0, 0)-triplet from the specified node. */
    packed_edges_plain(const key_type node_from) :
            data(node_from << node_shift) { }

    /* Construct a (node, base, bitset)-triplet from the specified key (node,
     * base) and bitset. */
    packed_edges_plain(const key_type key, const bitset_type bitset) :
            data((key << base_shift) | bitset) { }

    /* Construct a (node, base, bitset)-triplet from the specified node, base
     * and bitset. */
    packed_edges_plain(const node_type node_from, const node_type node_base, const node_type bitset) :
            data((node_from << node_shift) | (node_base << base_shift) | bitset) { }

    /* Return the (node, base, children)-pair representing the single edge
     * (node_from, node_to). */
    static pair_type create_edge(const node_type node_from, const node_type node_to) {
        return pair_type(node_from, node_to / packing_factor, one << (node_to % packing_factor));
    }


    /* Get the from-node. */
    node_type node_from() const {
        return data >> node_shift;
    }

    /* Get the base bitpattern. */
    node_type nodes_to_base() const {
        return (data >> base_shift) & base_mask;
    }


    /* Get the base. */
    node_type nodes_to_base_pf() const {
        return nodes_to_base() * packing_factor;
    }

    /* Get the key (node, base). */
    node_type node_key() const {
        return data >> base_shift;
    }

    /* Get the bitset. */
    bitset_type nodes_to_field() const {
        return data & bitset_mask;
    }

    /* Add bits (specified by the bitset set) to the bitset. */
    void add_bits_to_bitset(const bitset_type set) {
        data |= set;
    }


    /* Provide a strict ordering on (node, children)-pairs. */
    bool operator<(const pair_type& other) const {
        return data < other.data;
    }

    /* Return true if the other (node, children)-pairs is equivalent to this
     * (node, children)-pair. */
    bool operator==(const pair_type& other) const {
        return data == other.data;
    }


    /* key_bitset_pair interface get_key-function, as used by the
     * key_bitset-algorithms. */
    static node_type get_key(const pair_type& entry) {
        return entry.node_key();
    }

    /* key_bitset_pair interface get_bitset-function, as used by the
     * key_bitset-algorithms. */
    static node_type get_bitset(const pair_type& entry) {
        return entry.nodes_to_field();
    }

private:
    /* Parameters used for the bitwise implementation: in the (node, base,
     * bitset)-triple, the bitset are the first packing factor-bits, the
     * remaining bits are used to store the base and the node; in which node 
     * constitutes the most significant bits. The node_shift and base_shift
     * provides the number of bits the node and base are shifted and the
     * base_mask and bitset_mask provides bitwise masks used to filter out a
     * the base and the bitset from the underlying unsigned integer. */
    static constexpr node_type base_shift = packing_factor;
    static constexpr node_type node_shift = base_size + base_shift;
    static constexpr node_type base_mask = (one << base_size) - 1;

    node_type data;
};

/* Pair representation of packed edges. */
template<class NodeType, std::size_t PackingFactor>
class relation::details::packed_edges_pair {
public:
    typedef NodeType node_type;
    typedef node_type key_type;
    typedef node_type bitset_type;

    static constexpr bitset_type packing_factor = PackingFactor;
    typedef packed_edges_traits<packing_factor> traits;
    typedef packed_edges_pair<node_type, packing_factor> pair_type;

    static constexpr node_type one = 1;
    static constexpr node_type digits = std::numeric_limits<node_type>::digits;
    static constexpr node_type key_size = digits;
    static constexpr node_type node_size = (key_size / 2) + traits::added_bits;
    static constexpr node_type base_size = (key_size / 2) - traits::added_bits;
    static constexpr bitset_type bitset_mask = traits::bitset_mask;

    static constexpr node_type max_num_nodes = one << node_size;

    /* Construct an empty (node, base, bitset)-triplet. */
    packed_edges_pair() : key(0), bitset(0) { }

    /* Construct a (node, 0, 0)-triplet from the specified node. */
    packed_edges_pair(const key_type node_from) :
            key(node_from << node_shift), bitset(0) { }

    /* Construct a (node, base, bitset)-triplet from the specified key (node,
     * base) and bitset. */
    packed_edges_pair(const key_type key, const bitset_type bitset) :
            key(key), bitset(bitset) { }

    /* Construct a (node, base, bitset)-triplet from the specified node, base
     * and bitset. */
    packed_edges_pair(const node_type node_from, const node_type node_base, const bitset_type bitset) :
            key((node_from << node_shift) | node_base), bitset(bitset) { }

    /* Return the (node, base, children)-pair representing the single edge
     * (node_from, node_to). */
    static pair_type create_edge(const node_type node_from, const node_type node_to) {
        return pair_type(node_from, node_to / packing_factor, one << (node_to % packing_factor));
    }


    /* Get the from-node. */
    node_type node_from() const {
        return key >> node_shift;
    }

    /* Get the base bitpattern. */
    node_type nodes_to_base() const {
        return key & base_mask;
    }

    /* Get the base . */
    node_type nodes_to_base_pf() const {
        return nopdes_to_base() * packing_factor;
    }

    /* Get the key (node, base). */
    node_type node_key() const {
        return key;
    }

    /* Get the bitset. */
    bitset_type nodes_to_field() const {
        return bitset;
    }

    /* Add bits (specified by the bitset set) to the bitset. */
    void add_bits_to_bitset(const bitset_type set) {
        bitset |= set;
    }


    /* Provide a strict ordering on (node, children)-pairs. */
    bool operator<(const pair_type& other) const {
        return (key < other.key) || ((key == other.key) && (bitset < other.bitset));
    }

    /* Return true if the other (node, children)-pairs is equivalent to this
     * (node, children)-pair. */
    bool operator==(const pair_type& other) const {
        return (key == other.key) && (bitset == other.bitset);
    }


    /* key_bitset_pair interface get_key-function, as used by the
     * key_bitset-algorithms. */
    static node_type get_key(const pair_type& entry) {
        return entry.node_key();
    }

    /* key_bitset_pair interface get_bitset-function, as used by the
     * key_bitset-algorithms. */
    static node_type get_bitset(const pair_type& entry) {
        return entry.nodes_to_field();
    }

private:
    /* Parameters used for the bitwise implementation. The node and base are
     * encoded in a single value, in which the node occupies the most
     * significant bits.  */
    static constexpr node_type node_shift = base_size;
    static constexpr node_type base_mask = (one << base_size) - 1;

    key_type key;
    bitset_type bitset;
};

/* Compact edge-list implementation of a relation: we maintain an list of
 * edge-entries, in which each edge-entry represents a set of edges. The list of
 * edges is ordered on the from-node. This implementation relies on a buffer
 *  whose size has a quadratic upper bounded on the number of nodes. */
template<class NodeType, const std::size_t PackingFactor, bool PairRepresentation>
class relation::packed_edges_list {
private:
    typedef NodeType node_type;
    static constexpr std::size_t packing_factor = PackingFactor;
    static constexpr bool pair_representation = PairRepresentation;
    typedef typename std::conditional<pair_representation,
                                      details::packed_edges_pair<node_type, packing_factor>,
                                      details::packed_edges_plain<node_type, packing_factor>>::type packed_edge;
    typedef array<packed_edge> internal_edge_list;

public:
    typedef packed_edges_list<node_type, packing_factor, pair_representation> relation_type;
    typedef internal_edge_list buffer_type;
    static constexpr std::size_t max_num_nodes = packed_edge::max_num_nodes;

    packed_edges_list() : edges() { }

    packed_edges_list(const relation_type& other) : edges(other.edges) { }

    packed_edges_list(relation_type&& other) : edges(std::move(other.edges)) { }

    static buffer_type initial_buffer(const std::size_t max_num_nodes) {
        auto value = overflow_multiply<std::size_t>(2, max_num_nodes, max_num_nodes);
        if (value.second) {
            throw std::bad_alloc();
        }
        return buffer_type(value.first);
    }

    bool operator<(const relation_type& other) const {
        return edges.size() < other.edges.size() ||
                (edges.size() == other.edges.size() && edges < other.edges);
    }

    bool is_empty() const {
        return edges.size() == 0;
    }

    static relation_type relation(const raw_relation& relation, buffer_type& buffer) {
        if (relation.empty()) {
            return relation_type();
        }

        auto r_it = relation.begin(), r_end = relation.end();
        auto it = buffer.begin();
        *it = packed_edge::create_edge(r_it->from, r_it->to);
        ++r_it;

        auto key = it->node_key();
        while (r_it != r_end) {
            auto entry = packed_edge::create_edge(r_it->from, r_it->to);
            if (entry.node_key() == key) {
                it->add_bits_to_bitset(entry.nodes_to_field());
            }
            else {
                ++it;
                *it = entry;
                key = entry.node_key();
            }
            ++r_it;
        }
        ++it;
        return relation_type(buffer.begin(), it);
    }

    raw_relation to_relation() const {
        raw_relation result;
        for (auto edge_entry : edges) {
            auto node_from = edge_entry.node_from();
            auto node_to_base = edge_entry.nodes_to_base_pf();
            for (auto node_to : make_bit_iterator(edge_entry.nodes_to_field(), node_to_base)) {
                result.emplace_back(static_cast<node_id_type>(node_from),
                                    static_cast<node_id_type>(node_to));
            }
        }
        return result;
    }

    static relation_type identity(const std::size_t n, buffer_type& buffer) {
        auto it = buffer.begin();
        for (node_type node = 0; node < n; ++node) {
            *it = packed_edge::create_edge(node, node);
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type diversity(const std::size_t n, buffer_type& buffer) {
        auto it = buffer.begin();
        node_type full_blocks = static_cast<node_type>(n / packing_factor);
        node_type num_remainder = static_cast<node_type>(n % packing_factor);
        for (node_type node = 0; node < n; ++node) {
            auto base_it = it;
            for (node_type to_base = 0; to_base < full_blocks; ++to_base) {
                *it = packed_edge(node, to_base, packed_edge::bitset_mask);
                ++it;
            }
            if (num_remainder != 0) {
                *it = packed_edge(node, full_blocks, (packed_edge::one << num_remainder) - 1);
                ++it;
            }

            packed_edge identity = packed_edge::create_edge(node, node);
            packed_edge& target = base_it[identity.nodes_to_base()];
            target = packed_edge(node, identity.nodes_to_base(),
                                 target.nodes_to_field() & ~identity.nodes_to_field());
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_converse(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto& entry : relation.edges) {
            auto node_from = entry.node_from();
            for (auto node_to : make_bit_iterator(entry.nodes_to_field(), entry.nodes_to_base_pf())) {
                *it = packed_edge::create_edge(node_to, node_from);
                ++it;
            }
        }
        std::sort(buffer.begin(), it);
        return relation_type(buffer.begin(), pack_buffer(buffer.begin(), it));
    }

    static relation_type relation_project_1(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        auto rit = relation.edges.begin(), rend = relation.edges.end();
        if (rit != rend) {
            *it = packed_edge::create_edge(rit->node_from(), rit->node_from());
            ++rit;
            while (rit != rend) {
                if (it->node_from() != rit->node_from()) {
                    ++it;
                    *it = packed_edge::create_edge(rit->node_from(), rit->node_from());
                }
                ++rit;
            }
            ++it;
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_project_2(const relation_type& relation, buffer_type& buffer) {
        auto it = buffer.begin();
        for (auto& entry : relation.edges) {
            for (auto node_to : make_bit_iterator(entry.nodes_to_field(), entry.nodes_to_base_pf())) {
                *it = packed_edge::create_edge(node_to, node_to);
                ++it;
            }
        }
        std::sort(buffer.begin(), it);
        return relation_type(buffer.begin(), pack_buffer(buffer.begin(), it));
    }

    static relation_type relation_compose(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto it = buffer.begin();
        auto lhs_it = lhs.edges.begin();
        while (lhs_it != lhs.edges.end()) {
            auto it_begin = it;
            auto node_from = lhs_it->node_from();
            auto rhs_it = rhs.edges.begin();

            do {
                for (auto node_to : make_bit_iterator(lhs_it->nodes_to_field(), lhs_it->nodes_to_base_pf())) {
                    rhs_it = std::lower_bound(rhs_it, rhs.edges.end(), packed_edge(node_to));
                    while (rhs_it != rhs.edges.end() && rhs_it->node_from() == node_to) {
                        *it = packed_edge(node_from, rhs_it->nodes_to_base(), rhs_it->nodes_to_field());
                        ++it;
                        ++rhs_it;
                    }
                }
                ++lhs_it;
            } while (lhs_it != lhs.edges.end() && lhs_it->node_from() == node_from);
            std::sort(it_begin, it);
            it = pack_buffer(it_begin, it);
        }
        return relation_type(buffer.begin(), it);
    }

    static relation_type relation_union(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = details::key_bitset_union(lhs.edges.begin(), lhs.edges.end(),
                                             rhs.edges.begin(), rhs.edges.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

    static relation_type relation_intersect(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = details::key_bitset_intersection(lhs.edges.begin(), lhs.edges.end(),
                                                    rhs.edges.begin(), rhs.edges.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

    static relation_type relation_difference(const relation_type& lhs, const relation_type& rhs, buffer_type& buffer) {
        auto end = details::key_bitset_difference(lhs.edges.begin(), lhs.edges.end(),
                                                  rhs.edges.begin(), rhs.edges.end(), buffer.begin());
        return relation_type(buffer.begin(), end);
    }

private:
    internal_edge_list edges;

    /* Construct a packed edge list from a sorted list of packed edges. */
    template<class InputIterator>
    packed_edges_list(InputIterator first, InputIterator last) : edges(first, last) { }

    /* Make a sorted list of packed edge entries unique such that the (n, b)-
     * pairs are unique identifiers for each entry; this by merging consecutive
     * entries for the same (n, b)-pair. Return the iterator of the
     * one-past-last entry in the resulting list. */
    template<class InputIterator>
    static InputIterator pack_buffer(InputIterator first, InputIterator last) {
        if (first == last) {
            return first;
        }

        auto it_base = first;
        auto it = it_base + 1;
	    while (it != last) {
            auto key = it_base->node_key();
            while (it != last && key == it->node_key()) {
                it_base->add_bits_to_bitset(it->nodes_to_field());
                ++it;
            }
            if (it != last) {
                ++it_base;
                *it_base = *it;
                ++it;
            }
        }
	    return it_base + 1;
    }
};
