/**
 * Copyright (c) 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* We need AVX support for the datatypes used by the matrix16 implementation
 * (and we need SSE2 and AVX support for the implementation of matrix8 and
 * matrix16). */
#include "immintrin.h"

/* Low-level layout of standard int. */
union relation::details::field_int {
    int value;
    std::uint8_t parts_8[4];
    std::uint16_t parts_16[2];
    std::uint32_t parts_32[1];
};

/* Low-level layout of standard 64bit unsigned int. */
union relation::details::field_64 {
    std::uint64_t value;
    std::uint8_t parts_8[8];
    std::uint16_t parts_16[4];
    std::uint32_t parts_32[2];
    std::uint64_t parts_64[1];
};

/* Low-level layout of standard 128bit int SSE type. */
union relation::details::field_128 {
    __m128i value;
    std::uint8_t parts_8[16];
    std::uint16_t parts_16[8];
    std::uint32_t parts_32[4];
    std::uint64_t parts_64[2];
};

/* Low-level layout of standard 256bit int AVX type. */
union relation::details::field_256 {
    __m256i value;
    std::uint8_t parts_8[32];
    std::uint16_t parts_16[16];
    std::uint32_t parts_32[8];
    std::uint64_t parts_64[4];
};

/* A boolean matrix implementation of binary relations. This implementation
 * supports graph with at most 8 nodes. */
class relation::matrix8 {
public:
    typedef matrix8 relation_type;
    typedef relation_inplace_tag buffer_type;
    static constexpr std::size_t max_num_nodes = 8;

    matrix8() : data(0) { }

    matrix8(const relation_type& other) : data(other.data) { }

    matrix8(relation_type&& other) : data(other.data) { }

    bool operator<(const relation_type& other) const {
        return data < other.data;
    }

    bool is_empty() const {
        return data == 0;
    }

    static relation_type relation(const raw_relation& relation) {
        matrix8 result;
        for (auto edge : relation) {
            result.parts[edge.from] |= (one << edge.to);
        }
        return result;
    }

    raw_relation to_relation() const {
        raw_relation result;
        for (std::uint16_t i = 0; i < 8; ++i) {
            for (auto to : make_bit_iterator(parts[i])) {
                result.emplace_back(i, to);
            }
        }
        return result;
    }

    static relation_type identity(const std::size_t n) {
        constexpr std::uint64_t values[] = {
                0x0000000000000000, /* 0 */
                0x0000000000000001, /* 1 */
                0x0000000000000201, /* 2 */
                0x0000000000040201, /* 3 */
                0x0000000008040201, /* 4 */
                0x0000001008040201, /* 5 */
                0x0000201008040201, /* 6 */
                0x0040201008040201, /* 7 */
                0x8040201008040201  /* 8 */
        };
        return values[n];
    }

    static relation_type diversity(const std::size_t n) {
        constexpr std::uint64_t values[] = {
                0x0000000000000000, /* 0 */
                0x0000000000000000, /* 1 */
                0x0000000000000102, /* 2 */
                0x0000000000030506, /* 3 */
                0x00000000070B0D0E, /* 4 */
                0x0000000F171B1D1E, /* 5 */
                0x00001F2F373B3D3E, /* 6 */
                0x003F5F6F777B7D7E, /* 7 */
                0x7FBFDFEFF7FBFDFE  /* 8 */
        };
        return values[n];
    }

    static relation_type relation_converse(const relation_type& relation) {
        details::field_64 t;
        __m128i input = _mm_cvtsi64_si128(relation.data);

        /* Using movemask, we get the most-significant bit in each byte, hence, we
         * read the columns in reverse order (and set the rows in reverse order). */
        for (auto i = 7; i > 0; --i) {
            t.parts_8[i] = _mm_movemask_epi8(input);
            input = _mm_slli_epi16(input, 1);
        }

        /* Unroll for i = 0, saves a bitshift, slightly faster in practice. */
        t.parts_8[0] = _mm_movemask_epi8(input);
        return t.value;
    }

    static relation_type relation_project_1(const relation_type& relation) {
        details::field_64 source { relation.data };
        details::field_64 target;
        for (auto i = 0; i < 8; ++i) {
            target.parts_8[i] = (source.parts_8[i] != 0) << i;
        }
        return target.value;
    }

    static relation_type relation_project_2(const relation_type& relation) {
        return relation_project_1(relation_converse(relation));
    }

    static relation_type relation_compose(const relation_type& lhs, const relation_type& rhs) {
        matrix8 trhs = relation_converse(rhs);
        __m128i rhs_matrix = _mm_cvtsi64_si128(trhs.data);
        __m128i zero = _mm_setzero_si128();

        details::field_64 target;
        details::field_64 source { lhs.data };
        for (auto i = 0; i < 8; ++i) {
            /* Compute product of i-th row times each column (each row in the
                * transpose). */
            __m128i row = _mm_set1_epi8(source.parts_8[i]);
            __m128i product = _mm_and_si128(row, rhs_matrix);

            /* If the j-th byte is all-zero, then the product of i-th row and the
                * j-th column is zero. */
            __m128i zerod_bytes = _mm_cmpeq_epi8(product, zero);
            target.parts_8[i] = ~(_mm_movemask_epi8(zerod_bytes));
        }

        return target.value;
    }

    static relation_type relation_union(const relation_type& lhs, const relation_type& rhs) {
        return lhs.data | rhs.data;
    }

    static relation_type relation_intersect(const relation_type& lhs, const relation_type& rhs) {
        return lhs.data & rhs.data;
    }

    static relation_type relation_difference(const relation_type& lhs, const relation_type& rhs) {
        return lhs.data & ~rhs.data;
    }

private:
    static constexpr std::size_t one = 1;

    /* Construct a boolean matrix from the internal representation. */
    matrix8(const std::uint64_t data) : data(data) { }

    /* Internally, the boolean matrix is an 64bit (8 by 8) unsigned integer. We
     * use the union to easily access individual bytes in this 64bit unsigned
     * integer (which represent the rows of the matrix). */
    union {
        std::uint64_t data;
        std::uint8_t parts[8];
    };
};

/* A boolean matrix implementation of binary relations. This implementation
 * supports graph with at most 126 nodes. */
class relation::matrix16 {
public:
    typedef matrix16 relation_type;
    typedef relation_inplace_tag buffer_type;
    static constexpr std::size_t max_num_nodes = 8;

    matrix16() : data(_mm256_setzero_si256()) { }

    matrix16(const relation_type& other) : data(other.data) { }

    matrix16(relation_type&& other) : data(other.data) { }

    bool operator<(const relation_type& other) const {
        return parts < other.parts; 
    }

    bool is_empty() const {
        return _mm256_testz_si256(data, data) == 0;
    }

    static relation_type relation(const raw_relation& relation) {
        matrix16 result;
        for (auto edge : relation) {
            std::size_t to_shift = edge.to >> 3;
            result.byte_parts[edge.from + to_shift * 16] |=
                    (one << (edge.to - to_shift * 8));
        }
        return result;
    }

    raw_relation to_relation() const {
        raw_relation result;
        for (std::uint16_t i = 0; i < 16; ++i) {
            std::uint16_t row = byte_parts[i] + (byte_parts[i + 16] << 8);
            for (auto to : make_bit_iterator(row)) {
                result.emplace_back(i, to);
            }
        }
        return result;
    }

    static relation_type identity(const std::size_t n) {
        constexpr std::uint64_t values[][2] = {
                {0x0000000000000000, 0x0000000000000000}, /* 0 */
                {0x0000000000000000, 0x0000000000000001}, /* 1 */
                {0x0000000000000000, 0x0000000000000201}, /* 2 */
                {0x0000000000000000, 0x0000000000040201}, /* 3 */
                {0x0000000000000000, 0x0000000008040201}, /* 4 */
                {0x0000000000000000, 0x0000001008040201}, /* 5 */
                {0x0000000000000000, 0x0000201008040201}, /* 6 */
                {0x0000000000000000, 0x0040201008040201}, /* 7 */
                {0x0000000000000000, 0x8040201008040201}, /* 8 */
                {0x0000000000000001, 0x8040201008040201}, /* 9 */
                {0x0000000000000201, 0x8040201008040201}, /* 10 */
                {0x0000000000040201, 0x8040201008040201}, /* 11 */
                {0x0000000008040201, 0x8040201008040201}, /* 12 */
                {0x0000001008040201, 0x8040201008040201}, /* 13 */
                {0x0000201008040201, 0x8040201008040201}, /* 14 */
                {0x0040201008040201, 0x8040201008040201}, /* 15 */
                {0x8040201008040201, 0x8040201008040201}, /* 16 */
        };
        return _mm256_set_epi64x(values[n][0], 0, 0, values[n][1]);
    }

    static relation_type diversity(const std::size_t n) {
        constexpr std::uint64_t values[][4] = {
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x0000000000000000}, /* 0 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x0000000000000000}, /* 1 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x0000000000000102}, /* 2 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x0000000000030506}, /* 3 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x00000000070B0D0E}, /* 4 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x0000000F171B1D1E}, /* 5 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x00001F2F373B3D3E}, /* 6 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x003F5F6F777B7D7E}, /* 7 */
                {0x0000000000000000, 0x0000000000000000, 0x0000000000000000, 0x7FBFDFEFF7FBFDFE}, /* 8 */
                {0x0000000000000000, 0x0101010101010101, 0x00000000000000FF, 0x7FBFDFEFF7FBFDFE}, /* 9 */
                {0x0000000000000102, 0x0303030303030303, 0x000000000000FFFF, 0x7FBFDFEFF7FBFDFE}, /* 10 */
                {0x0000000000030506, 0x0707070707070707, 0x0000000000FFFFFF, 0x7FBFDFEFF7FBFDFE}, /* 11 */
                {0x00000000070B0D0E, 0x0F0F0F0F0F0F0F0F, 0x00000000FFFFFFFF, 0x7FBFDFEFF7FBFDFE}, /* 12 */
                {0x0000000F171B1D1E, 0x1F1F1F1F1F1F1F1F, 0x000000FFFFFFFFFF, 0x7FBFDFEFF7FBFDFE}, /* 13 */
                {0x00001F2F373B3D3E, 0x3F3F3F3F3F3F3F3F, 0x0000FFFFFFFFFFFF, 0x7FBFDFEFF7FBFDFE}, /* 14 */
                {0x003F5F6F777B7D7E, 0x7F7F7F7F7F7F7F7F, 0x00FFFFFFFFFFFFFF, 0x7FBFDFEFF7FBFDFE}, /* 15 */
                {0x7FBFDFEFF7FBFDFE, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF, 0x7FBFDFEFF7FBFDFE}  /* 16 */
        };
        return _mm256_set_epi64x(values[n][0], values[n][1], values[n][2], values[n][3]);
    }

    static relation_type relation_converse(const relation_type& relation) {
        details::field_256 t;

        /* Using movemask, we get the most-significant bit in each byte, hence,
         * we read the partial columns in reverse order. The 32bit result
         * consists of two full 16bit columns, namely the i-th column and the
         * (i+8)-th column (which will become the i-th and (i+8)-th row).
         *  Observe that the j-th 16bit row needs to be stored in the j-th byte
         * (first 8 columns) and the (j+16)-th byte (last 8 columns). */

        /* Unroll loop for case i = 7, save unnecessary copy. */
        details::field_int result { _mm256_movemask_epi8(relation.data) };
        __m256i source = _mm256_slli_epi16(relation.data, 1);
        t.parts_8[7] = result.parts_8[0];
        t.parts_8[15] = result.parts_8[2];
        t.parts_8[23] = result.parts_8[1];
        t.parts_8[31] = result.parts_8[3];

        /* Normal loop for case i = 6 ... 1; unrolled cases are +-2% faster. */
        for (auto i = 6; i > 0; --i) {
            result.value = _mm256_movemask_epi8(source);
            source = _mm256_slli_epi16(source, 1);
            t.parts_8[i] = result.parts_8[0];
            t.parts_8[i + 8] = result.parts_8[2];
            t.parts_8[i + 16] = result.parts_8[1];
            t.parts_8[i + 24] = result.parts_8[3];
        }

        /* Unroll loop for case i = 0, save unnecessary shift. */
        result.value = _mm256_movemask_epi8(source);
        t.parts_8[0] = result.parts_8[0];
        t.parts_8[8] = result.parts_8[2];
        t.parts_8[16] = result.parts_8[1];
        t.parts_8[24] = result.parts_8[3];
       
        return t.value;    
    }

    static relation_type relation_project_1(const relation_type& relation) {
        __m128i low = _mm256_extractf128_si256(relation.data, 0);
        __m128i high = _mm256_extractf128_si256(relation.data, 1);
        details::field_128 source { _mm_or_si128(low, high) };
        details::field_256 target;
        for (auto i = 0; i < 8; ++i) {
            target.parts_8[i] = (source.parts_8[i]  != 0) << i;
            target.parts_8[i + 16] = 0;
        }
        for (auto i = 0; i < 8; ++i) {
            target.parts_8[i + 24] = (source.parts_8[i + 8]  != 0) << i;
            target.parts_8[i + 8] = 0;
        }
        return target.value;
    }

    static relation_type relation_project_2(const relation_type& relation) {
        return relation_project_1(relation_converse(relation));
    }

    static relation_type relation_compose(const relation_type& lhs, const relation_type& rhs)  {
        matrix16 tother = relation_converse(rhs);
        __m256i zero = _mm256_setzero_si256();

        details::field_256 source { lhs.data };
        details::field_256 target;
        for (std::size_t i = 0; i < 16; ++i) {
            /* Compute product of i-th row times each column (each row in the
             * transpose). Observe that rows are split in a low/high part: the
             * first 128 bit of a matrix correspond with the first 8 columns of
             * the 16 rows, the last 128 bit of a matrix correspond with the
             * last 8 columns of the 16 rows. Hence: we also need to
             * redistribute our original row representation. */
            __m256i row = _mm256_set_m128i(_mm_set1_epi8(source.parts_8[i + 16]),
                                            _mm_set1_epi8(source.parts_8[i]));
            __m256i product = _mm256_and_si256(row, tother.data);

            /* If the j-th byte is all-zero (0 <= j < 16), then the product of
             * the first 8 columns of the i-th row and the first 8 rows of the 
             * j-th column is zero. If the j-th byte is all-zero (16 <= j < 32),
             * then the product of the last 8 columns of the i-th row and the
             * last 8 rows of the j-th column is zero. */
            __m256i zerod_bytes = _mm256_cmpeq_epi8(product, zero);
            details::field_int result { ~(_mm256_movemask_epi8(zerod_bytes)) };
            target.parts_8[i] = result.parts_8[0] | result.parts_8[2];
            target.parts_8[i + 16] = result.parts_8[1] | result.parts_8[3];
        }

        return target.value;
    }

    static relation_type relation_union(const relation_type& lhs, const relation_type& rhs) {
        return _mm256_or_si256(lhs.data, rhs.data);
    }

    static relation_type relation_intersect(const relation_type& lhs, const relation_type& rhs) {
        return _mm256_and_si256(lhs.data,  rhs.data);
    }

    static relation_type relation_difference(const relation_type& lhs, const relation_type& rhs) {
        return _mm256_andnot_si256(rhs.data, lhs.data);
    }

private:
    static constexpr std::size_t one = 1;

    /* Construct a boolean matrix from the internal representation. */
    matrix16(const __m256i data)  : data(data) { }

    /* Internally, the boolean matrix is an 256bit (16 by 16) unsigned integer.
     * We use __m256i (Intel AVX), and we use the union to easily access
     * individual bytes in this 256bit unsigned integer (which represent
     * half-rows of the matrix). */
    union {
        __m256i data;
        std::array<std::uint64_t, 4> parts;
        std::uint8_t byte_parts[32];
    };
};