/**
 * Copyright (c) 2014, 2016 Jelle Hellings.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JELLE HELLINGS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE FOUNDATION OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef INCLUDE_STRING_PARSER_HPP_
#define INCLUDE_STRING_PARSER_HPP_
#include <algorithm>
#include <string>
#include <utility>
#include "optional.hpp"

/* Simple iterator-through-string provider with functions to accommodate
 * backtracking and functions to perform pretty-printed custom errors and
 * warnings (including line and column counting). */
class string_reader {
public:
    /* The type of characters, strings, and positions used by this reader. */
    typedef char char_type;
    typedef std::basic_string<char_type> string_type;
    typedef string_type::size_type size_type;


    /* Structure to represent information on a position in the source. */
    struct position_info {
        /* Construct a new position information record. */
        position_info(size_type line, size_type column, size_type start_line) :
                line(line), column(column), start_line(start_line) { }


        /* The calculated information: line number, column number in the line,
         * and the position where the line starts */
        const size_type line;
        const size_type column;
        const size_type start_line;
    };

    /* Construct a string reader for the specified source. */
    string_reader(const std::string& source) : pos(0), source(source)  { }

    /* Move-constructor. */
    string_reader(string_reader&& other) : pos(other.pos),
                                           source(std::move(other.source)) { }

    /* Destructor. */
    virtual ~string_reader() { }


    /* Return true if we are at the end. */
    bool end() const {
        return pos >= length();
    }

    /* Return the length of the string. */
    size_type length() const {
        return source.length();
    }

    /* Move to the next position. */
    void next() {
        ++pos;
    }

    /* Read the current character (pre: !end(), std::runtime_error). */
    char_type peek_char() const {
        if (end()) {
            throw_error("reading at end-of-string");
        }
        return source[pos];
    }

    /* Return the current position. */
    size_type position() const {
        return pos;
    }

    /* Set the current position. */
    void set_position(size_type new_position) {
        pos = new_position;
    }

    /* Return a part of the source starting at first, ending at last (excluding
     * last) (pre:0 <= first <= last <= length(), std::runtime_error). */
    string_type slice(size_type first, size_type last) const {
        if (!(0 <= first && first <= last && last <= length())) {
            throw_error("invalid slice: [" + std::to_string(first) + "," +
                                             std::to_string(last) +  "]");
        }
        return source.substr(first, last - first);
    }


    /* Return information on the specified position in the source. */
    position_info query_position(size_type position) const;

    /* Print a message to std::cerr, and stop with execution (an exception is
     * thrown). */
    void throw_error(const std::string& message) const;

private:
    size_type pos;
    string_type source;
};

/* The parser namespace provides functionality to quickly build simple
 * syntax-rule-based string parsers on top of the string_reader. */
namespace parser {
    template<class CapFn, class ValueType> class capture_class_implementation;

    /* A simple string parser that executes parse rules (see parse()). Use
     * the create() function for creating a parser. */
    template<class SkipFn>
    class parser_class_implementation : public string_reader {
    public:
        /* Construct a simple string parser using the provided string as a
         * source and the provided whitespace/comment-skip-function. */
        parser_class_implementation(const std::string& source, SkipFn skip_function) :
                string_reader(source), skip_function(skip_function) { }

        /* Move-constructor. */
        parser_class_implementation(parser_class_implementation<SkipFn>&& other) :
                string_reader(std::move(other.string_reader)), skip_function(other.skip_function) { }


        /* Rule-based parser: the arguments are evaluated from left to right.
         * Individual arguments are evaluated by the read_atom() functions. On
         * successful parsing, true is returned. If any of the arguments fail,
         * the original situation is restored and false is returned. Any
         * whitespace or comments, as recognized by the provided skip function,
         * are greedily skipped over before each argument is evaluated. */
        template<class... Args> bool parse(Args... args) {
            auto pos = position();
            if (read_atoms(args...)) {
                skip_function(*this);
                return true;
            }
            set_position(pos);
            return false;
        }

    private:
        /* Apply each argument in sequence, left-to-right. */
        template<class Atom, class... Args> bool read_atoms(Atom& value, Args... args) {
            return read_atom(value) && read_atoms(args...);
        }

        /* Base-case for read_atoms template. */
        template<class Atom> bool read_atoms(Atom& value) {
            return read_atom(value);
        }


        /* Read a string constant: check if the current position in the input
         * contains the specified string (first skip whitespace/comments).
         * Return true on success, else return false (the position in the source
         * might change) */
        bool read_atom(const std::string& value) {
            skip_function(*this);
            for (char ch : value) {
                if (end() || peek_char() != ch) {
                    return false;
                }
                next();
            }
            return true;
        }

        /* Read a capture class: execute the capture function at the current
         * reader state (first skip whitespace/comments), and if the function
         * returns a result, then set this result to the capture variable and
         * return true. Else return false, and the capture variable is
         * unchanged. */
        template<class CapFn, class ValueType>
        bool read_atom(capture_class_implementation<CapFn, ValueType>& capture) {
            skip_function(*this);
            auto value = capture.function(*this);
            if (value.has_value()) {
                std::swap(capture.target, value.get_value());
                return true;
            }
            return false;
        }

        /* Function used to skip whitespace and comments with. */
        SkipFn skip_function;
    };

    /* A capture class provides a parse function to evaluate a string reader,
     * and a reference to the variable to which the result of the parse function
     * should be written. Capture functions can easily be constructed using
     * parser::read_function(). Use the capture() function for creating a
     * parser. */
    template<class CapFn, class ValueType> class capture_class_implementation final {
    public:
        typedef ValueType value_type;
        typedef CapFn capture_function_type; 

        /* Construct a capture class for the provided target variable and
         * capture function. */
        capture_class_implementation(std::string& target, capture_function_type function) :
                target(target), function(function) { }

        /* Copy-constructor. */
        capture_class_implementation(const capture_class_implementation& other) :
                target(other.target), function(other.function) { }

    private:
        value_type& target;
        capture_function_type function;

        /* The method parser::read_atom() directly interacts with the capture
         * target and function. */
        template<class SkipFn> friend class parser_class_implementation;
    };


    /* The read function uses two character-predicates to greedily read tokens
     * from the current position in the string reader. The first character
     * read should match the first-predicate, the remaining characters should
     * match the rest-predicate. The standard std::isXXX functions in <cctype>
     * can be used as character-predicates. This function can be used to
     * construct capture functions: the function
     *     optional<std::string> read_qualified_name(string_reader& reader) {
     *         return read_function(reader, std::isalpha, std::isalnum);
     *     }
     * will, for example, read qualified names that start with a letter,
     * followed by letters and/or numbers. */
    template<class FnFirst, class FnRest>
    optional<std::string> read_function(string_reader& reader,
                                        FnFirst first, FnRest rest) {
        auto pos = reader.position();
        if (!reader.end() && first(reader.peek_char())) {
            do {
                reader.next();
            } while (!reader.end() && rest(reader.peek_char()));
            return optional<std::string>(reader.slice(pos, reader.position()));
        }
        return optional<std::string>();
    }

    /* Read function in which the first- and rest-predicates are equivalent. */
    template<class FnOnly>
    optional<std::string> read_function(string_reader& reader, FnOnly only) {
        return read_function(reader, only, only);
    }


    /* Create a string parser from a string source, using the specified skip
     * function. */
    template<class SkipFn>
    parser_class_implementation<SkipFn> create(const std::string& source, SkipFn function) {
        return parser_class_implementation<SkipFn>(source, function);
    }

    /* Create a capture class for the specified target variable (the capture
     * variable), using the specified capture function that optionally reads the
     * specified value type from a string reader. */
    template<class CapFn, class ValueType>
    capture_class_implementation<CapFn, ValueType> capture(ValueType& target, CapFn function) {
        return capture_class_implementation<CapFn, ValueType>(target, function);
    }
}

#endif